#!/usr/bin/env python -i
from __future__ import print_function

import re
from warnings import warn
from datetime import datetime

from PseudoNetCDF import PseudoNetCDFFile

#  20 lines reads hourly irr

_split1 = 'Environment Tables for'


def MorphoIRRt(irrpath):
    mrglines = open(irrpath).readlines()
    try:
        datelines = [l for l in mrglines if _split1 in l]
        datestr = datelines[0].split(_split1)[-1].strip()
        jday = int(datetime.strptime(datestr, '%d-%b-%y').strftime('%Y%j'))
    except Exception:
        warn('Could not find/parse date; using 1900001')
        jday = 1900001

    mrglines = [line for line in mrglines if line[:2]
                not in ('//', '**') and line not in ('', '\n')]
    name_line = mrglines.pop(0)
    irrlabel = re.compile(r'rt\[\S+\]')
    rxn_names = ['IRR_%s' % name.replace('rt[', '').replace(']', '')
                 for name in irrlabel.findall(name_line)]
    name_line = ['N', 'T'] + rxn_names
    unit_line = mrglines.pop(0).split()
    unit_line = [unit for unit in unit_line]
    assert(all([eval(v) == 0. for v in mrglines.pop(0).split()][2:]))
    mrgfile = PseudoNetCDFFile()
    mrgfile.createDimension('TSTEP', len(mrglines))
    mrgfile.createDimension('DATE-TIME', 2)
    mrgfile.createDimension('VAR', 1)
    unit_dict = dict([(k, v) for k, v in zip(name_line, unit_line)])
    tflag = mrgfile.createVariable('TFLAG', 'f', ('TSTEP', 'VAR', 'DATE-TIME'))
    tflag.units = '<JDAY, MIN>'
    tflag.long_name = tflag.var_desc = 'TFLAG'
    tflag[:, :, 0] = jday
    for name in name_line:
        var = mrgfile.createVariable(name, 'f', ('TSTEP',))
        var.units = unit_dict[name]
        var.long_name = var.var_desc = name
    for ti, line in enumerate(mrglines):
        for var_name, value in zip(name_line, line.split()):
            var = mrgfile.variables[var_name]
            var[ti] = float(value)

    for name in name_line:
        if name in ('T', 'N'):
            continue
        var = mrgfile.variables[name]
        var[1:] = (var[1:] - var[:-1]) * 1000.
    tflag[:, :, 1] = mrgfile.variables['T'][:, None]
    return mrgfile


def MorphoConc(concpath):
    conclines = open(concpath).readlines()
    conclines = [line for line in conclines if line[:2]
                 not in ('//', '**') and line not in ('', '\n')]
    name_line = conclines.pop(0)
    conclabel = re.compile(r'n\[\S+\]')
    conc_names = [name.replace('n[', '').replace(']', '')
                  for name in conclabel.findall(name_line)]
    name_line = ['N', 'T'] + conc_names
    unit_line = conclines.pop(0).split()
    unit_line = [unit for unit in unit_line]
    concfile = PseudoNetCDFFile()
    concfile.createDimension('TSTEP', len(conclines))
    concfile.createDimension('DATE-TIME', 2)
    concfile.createDimension('VAR', 1)
    unit_dict = dict([(k, v) for k, v in zip(name_line, unit_line)])
    for name in name_line:
        var = concfile.createVariable(name, 'f', ('TSTEP',))
        var.units = unit_dict[name]
        var.long_name = var.var_desc = name
    for ti, line in enumerate(conclines):
        for var_name, value in zip(name_line, line.split()):
            var = concfile.variables[var_name]
            var[ti] = float(value)

    return concfile


def MorphoPERMM(concpath, irrtpath):
    mrgf = MorphoIRRt(irrtpath)
    concf = MorphoConc(concpath)
    for key, var in concf.variables.items():
        initvar = mrgf.createVariable('INIT_%s' % key, 'f', ('TSTEP',))
        initvar.long_name = var.long_name
        initvar.var_desc = var.var_desc
        initvar.units = var.units
        initvar[:] = var[:-1]
        finalvar = mrgf.createVariable('FINAL_%s' % key, 'f', ('TSTEP',))
        finalvar.long_name = var.long_name
        finalvar.var_desc = var.var_desc
        finalvar.units = var.units
        finalvar[:] = var[1:]
    mrgf.Species = ' '.join([k.ljust(16) for k in concf.variables.keys()])
    mrgf.Processes = 'INIT'.ljust(17) + 'FINAL'.ljust(16)
    return mrgf


# if __name__ == '__main__':
#     from permm.guis.Simplewx import StartWx
#     import os
#     from permm.analyses.network.core import MakeCarbonTrace
#     from permm import Mechanism
#     concpath = sys.argv[1]
#     mrgpath = sys.argv[2]
#     irrtpath = sys.argv[3]
#     print(sys.argv[1:])
#     if os.path.exists('mech.yaml'):
#         cb05 = Mechanism('mech.yaml')
#     else:
#         cb05 = MorphoMrg(mrgpath, 'mech.yaml')
#     mrg = MorphoPERMM(concpath, irrtpath)
#     cb05.set_mrg(mrg)
#     cb05.globalize(globals())
#     n = MakeCarbonTrace(cb05, makes_larger=[PAR])
#     import networkx as nx
#     es = nx.dfs_edges(n)
#     #from permm.analyses.history import matrix
#     #history = matrix(cb05, [C2O3], [HC+Radical-OH-HO2-O1D], [])



from __future__ import print_function
__doc__ = r"""
PseudoNetCDF provides basic interfaces for emulating
NetCDF and manipulating or extending existing NetCDF
like files
"""

__all__ = ['aermodfiles',
           'anyfile',
           'ArrayTransforms',
           'camxfiles',
           'ceilometerfiles',
           'cmaqfiles',
           'conventions',
           'coordutil',
           'epafiles',
           'geoschemfiles',
           'icarttfiles',
           'MetaNetCDF',
           'noaafiles',
           'pnc',
           'PNC',
           'pncdump',
           'pncopen',
           'pncmfopen',
           'getreader',
           'getreaderdict',
           'pncwrite',
           'PseudoNetCDFFile',
           'PseudoNetCDFVariable',
           'PseudoNetCDFVariables',
           'racmfiles',
           'sci_var',
           'test',
           'textfiles',
           'toms',
           'units',
           'warn',
           'woudcfiles',
           'wrffiles']

from PseudoNetCDF.pncwarn import warn

from PseudoNetCDF.core import PseudoNetCDFFile
from PseudoNetCDF.core import PseudoNetCDFVariable, PseudoNetCDFVariables

from PseudoNetCDF import sci_var
from . import units
from . import test
from . import coordutil
from . import MetaNetCDF
from . import ArrayTransforms

from . import aermodfiles
from . import camxfiles
from . import ceilometerfiles
from . import cmaqfiles
from . import epafiles
from . import geoschemfiles
from . import icarttfiles
from . import noaafiles
from . import racmfiles
from . import textfiles
from . import toms
from . import woudcfiles
from . import wrffiles

from ._getreader import anyfile, pncopen, pncmfopen, getreader, getreaderdict
from ._getwriter import pncwrite
from .pncparse import PNC, pnc

for k in sci_var.__all__:
    if k not in __all__:
        __all__.append(k)
    globals()[k] = getattr(sci_var, k)


def makequite():
    global _quiet
    _quiet = True



from __future__ import print_function
__all__ = ['cdtoms', 'tomsl3']


from PseudoNetCDF import PseudoNetCDFFile
from PseudoNetCDF.pncwarn import warn
from re import compile
import numpy as np
from datetime import datetime


dayre = compile(' Day:\s+(?P<jday>\d+) (?P<daystring>.{12})\s+EP/' +
                'TOMS CORRECTED OZONE GEN:\d+\.\d+\sV\d ALECT:\s+\d+' +
                ':\d+ [AP]M ')
edgere = compile(
    '.+(?P<bins>\d+)\s+bins.*' +
    '\s+[-+]?(?P<start>\d+\.\d+)\s*(?P<startdir>[WENS]).*' +
    '\s+[-+]?(?P<end>\d+\.\d+)\s*(?P<enddir>[WENS])\s+.*' +
    '\((?P<step>\d+.\d+)\s*degree steps.+')


def _groupdict(reo, line):
    result = reo.match(line)
    if result is None:
        return {}
    else:
        return result.groupdict()


def cdtoms(path, outfile=None):
    if outfile is None:
        outfile = PseudoNetCDFFile()
    inlines = open(path, 'r').readlines()
    dayline = inlines[0]
    daygrp = _groupdict(dayre, dayline)

    sdate = 0
    if 'daystring' in daygrp:
        date = datetime.strptime(daygrp['daystring'], '%b %d, %Y')
        rdate = datetime(1970, 1, 1)
        sdate = (date - rdate).total_seconds()
    else:
        import pandas as pd
        dayparts = dayline.split(' ')
        for i in [3, 2, 1]:
            daystr = ' '.join(dayparts[:i])
            try:
                date = pd.to_datetime(daystr, box=False)
                break
            except Exception as e:
                print(e)
        else:
            date = np.datetime64('1970-01-01')
        rdate = np.datetime64('1970-01-01')
        sdate = (date - rdate).astype('d') / 1e9

    longrp = _groupdict(edgere, inlines[1])
    latgrp = _groupdict(edgere, inlines[2])

    for propdict in [daygrp, longrp, latgrp]:
        for k, v in propdict.items():
            try:
                v = eval(v)
            except Exception:
                pass
            setattr(outfile, k, v)
    outfile.HISTORY = ''.join(inlines[:3])

    blat = latgrp.get('start', '59.5')
    bsn = latgrp.get('startdir', 'S')
    elat = latgrp.get('end', '59.5')
    esn = latgrp.get('enddir', 'N')
    latstep = float(latgrp.get('step', '1'))
    blat = {'N': 1, 'S': -1}[bsn] * float(blat)
    elat = {'N': 1, 'S': -1}[esn] * float(elat)

    blon = longrp.get('start', '179.375')
    bwe = longrp.get('startdir', 'W')
    elon = longrp.get('end', '179.375')
    ewe = longrp.get('enddir', 'E')
    lonstep = float(longrp.get('step', '1.25'))
    blon = {'E': 1, 'W': -1}[bwe] * float(blon)
    elon = {'E': 1, 'W': -1}[ewe] * float(elon)

    datalines = inlines[3:]
    lats = []
    for i, line in enumerate(datalines):
        if 'lat' not in line:
            datalines[i] = line[1:-1].rstrip()
        else:
            data, lat = line.split('lat =')
            datalines[i] = data[1:-1].rstrip()
            lats.append(lat.strip())

    nlats = len(lats)
    datablock = ''.join(datalines).replace(' ', '0')
    nlons = len(datablock) // 3 // nlats
    outfile.createDimension('time', 1)
    outfile.createDimension('latitude', nlats)
    outfile.createDimension('longitude', nlons)
    outfile.createDimension('nv', 2)

    var = outfile.createVariable('time', 'f', ('time',))
    var.units = 'seconds since 1970-01-01 00:00:00+0000'
    var[:] = sdate

    var = outfile.createVariable('latitude', 'f', ('latitude',))
    var.units = 'degrees N'
    var[:] = np.arange(blat, elat + latstep, latstep)
    lat = var
    linelat = np.array(lats, dtype='f')
    if not (lat[:] == linelat).all():
        warn('Header metadata does not match lats')
        lat[:] = linelat

    var = outfile.createVariable('latitude_bounds', 'f', ('latitude', 'nv'))
    var.units = 'degrees N'
    var[:, 0] = lat - latstep / 2.
    var[:, 1] = lat + latstep / 2.

    var = outfile.createVariable('longitude', 'f', ('longitude',))
    var.units = 'degrees E'
    lon = var[:] = np.arange(blon, elon + lonstep, lonstep)

    var = outfile.createVariable('longitude_bounds', 'f', ('longitude', 'nv'))
    var.units = 'degrees E'
    var[:, 0] = lon - lonstep / 2.
    var[:, 1] = lon + lonstep / 2.

    var = outfile.createVariable(
        'ozone', 'f', ('latitude', 'longitude'),
        missing_value=999
    )
    var.units = 'matm-cm'
    var.long_name = var.var_desc = 'ozone'.ljust(16)
    var[:] = np.ma.masked_values(
        np.array(
            [i for i in datablock],
            dtype='S1'
        ).view('S3').astype('i'),
        var.missing_value
    ).reshape(nlats, nlons)

    return outfile


class tomsl3(PseudoNetCDFFile):
    def __init__(self, path):
        cdtoms(path, self)


if __name__ == '__main__':
    from PseudoNetCDF.pncdump import pncdump
    pfile = cdtoms('test.txt')
    pncdump(pfile)
    print(pfile.variables['ozone'][2, 3])



__doc__ = """
.. _irr
:mod:`irr` -- IRR File Interfaces
=================================

.. module:: irr
   :platform: Unix, Windows
   :synopsis: Provides :ref:`PseudoNetCDF` memory map and random access read
   based file interfaces for CAMx IRR files.
.. moduleauthor:: Barron Henderson <barronh@unc.edu>
"""
__all__ = ['Memmap', 'Read']

from . import Memmap
from . import Read

if __name__ == '__main__':
    from PseudoNetCDF.camxfiles.irr.Memmap import irr
    from PseudoNetCDF.pncdump import pncdump_parser, \
        dump_from_cmd_line
    parser = pncdump_parser()
    (file_path, options, extra_args_dict) = parser.parse_args()

    dump_from_cmd_line(file_path, options, irr)



import sys
import warnings

std_showwarning = warnings.showwarning


def warn(*args, **kwds):
    """
    Thin wrapper around warnings.warn to prepend **PNC:
    to warnings. This makes it easy to find PseudoNetCDF
    warnings.

    Arguments
    ---------
    args : arguments for warnings.warn
    kwds : keywords for warnings.warn

    Returns
    -------
    out : same as warnings.warn
    """

    warnings.showwarning = clean_showwarning
    out = warnings.warn(*args, **kwds)
    warnings.showwarning = std_showwarning
    return out


def clean_showwarning(message, category, filename, lineno, file=None,
                      line=None):
    global _first_read_only
    if file is None:
        file = sys.stderr
        if file is None:
            # sys.stderr is None when run with pythonw.exe - warnings get lost
            return
    try:
        warnstr = '**PNC:%s:%s:%s:\n  %s\n' % (
            (filename), lineno, category.__name__, message)
        file.write(warnstr)
    except OSError:
        pass  # the file (probably stderr) is invalid - this warning gets lost.
    return


filterwarnings = warnings.filterwarnings
simplefilter = warnings.simplefilter


def quiet():
    simplefilter('ignore', UserWarning)



__doc__ = """
.. _wind
:mod:`wind` -- Wind File Interfaces
===================================

.. module:: wind
   :platform: Unix, Windows
   :synopsis: Provides :ref:`PseudoNetCDF` memory map and random access read
   based file interfaces for CAMx wind files.
.. moduleauthor:: Barron Henderson <barronh@unc.edu>
"""
__all__ = ['Memmap', 'Read', 'Write', 'Transforms']

from . import Memmap
from . import Read
from . import Write
from . import Transforms

if __name__ == '__main__':
    from PseudoNetCDF.camxfiles.wind.Memmap import wind
    from PseudoNetCDF.pncdump import pncdump_parser, \
        dump_from_cmd_line
    parser = pncdump_parser()
    parser.add_argument("cols", int)
    parser.add_argument("rows", int)
    (file_path, options, extra_args_dict) = parser.parse_args()

    dump_from_cmd_line(file_path, options,
                       lambda path: wind(path, **extra_args_dict))



__all__ = ['plotts',
           'plotprofile',
           'add_vertprofile_options',
           'plotscatter',
           'SegmentedNorm',
           'SegmentedLogNorm',
           'plot',
           'plot2d',
           'plotcontour',
           'plotcontourf'
           ]

import pylab
import matplotlib.pyplot as pyplot
import numpy as np

from .pncts import plotts
from .vertprofile import plotprofile, add_vertprofile_options
from .pncscatter import pncscatter as plotscatter

pl = pylab
plt = pyplot

Normalize = pl.matplotlib.colors.Normalize
LogNorm = pl.matplotlib.colors.LogNorm
SymLogNorm = pl.matplotlib.colors.SymLogNorm
BoundaryNorm = pl.matplotlib.colors.BoundaryNorm


def SegmentedNorm(vmin, vmax, bins=10, ncolors=256):
    boundaries = np.linspace(vmin, vmax, bins + 1)
    return BoundaryNorm(boundaries, ncolors)


def SegmentedLogNorm(vmin, vmax, bins=10, ncolors=256, full_levels=False):
    """
    Create BoundaryNorm with (N=bins) log spacing.  If full_levels,
    then colorbar will start on a full log level and end on a full
    log level.



    boundaries = np.logspace(np.log10(vmin), np.log10(vmax), bins + 1)
    return BoundaryNorm(boundaries, ncolors)
    """
    if full_levels:
        boundaries = np.logspace(
            np.floor(np.log10(vmin)), np.ceil(np.log10(vmax)), bins + 1)
    else:
        boundaries = np.logspace(np.log10(vmin), np.log10(vmax), bins + 1)
    return BoundaryNorm(boundaries, ncolors)


LogFormatter = pl.matplotlib.ticker.LogFormatter
ScalarFormatter = pl.matplotlib.ticker.ScalarFormatter


def plot(args, plotfunc=plt.plot):
    for ifile in args.ifiles:
        for varkey in args.variables:
            y = ifile.variables[varkey]
            if args.squeeze:
                y = y.squeeze()
            plotfunc(y)


def plot2d(args):
    plot(args, plotfunc=plt.pcolor)


def plotcontour(args):
    plot(args, plotfunc=plt.contour)


def plotcontourf(args):
    plot(args, plotfunc=plt.contourf)



#!/usr/bin/env python
# Run this script with pnc options


def main():
    from PseudoNetCDF.pncparse import pncparse, getparser
    from PseudoNetCDF.plotutil.pnc2d import make2ds
    parser = getparser(has_ofile=True, plot_options=True, interactive=True)
    parser.add_argument('--swapaxes', action='store_true',
                        help='Swap x-y axes')
    ifiles, args = pncparse(
        has_ofile=True, plot_options=True, interactive=True, parser=parser)
    make2ds(args)


if __name__ == '__main__':
    main()



__all__ = ['mda8', 'daymax', 'daymin', 'daymean', 'daystd', 'dayvar']
import numpy as np


def mda8(arr, axis=None, keepdims=True):
    """
    Daily-Maximum 8-hour average concentration
       - can be applied to any dimensions, but make sense with time
       - returns the day max of the a8
    Arguments:
       arr - array like
       axis - axis over which to apply mda8
       keepdims - should be true
    Returns:
       out - maximum of 8 element running average. For masked inptus, only
             values with 6 or more valid entries are returned
    """
    if hasattr(arr, 'mask'):
        arr_m = np.ma.masked_array(arr)
        nvalid = np.apply_along_axis(
            np.convolve, axis=axis, arr=~np.ma.getarraymask(arr_m),
            v=[1] * 8, mode='valid')
        arra8_m = np.ma.masked_where(nvalid < 6, np.apply_along_axis(
            np.convolve, axis=axis, arr=arr_m.filled(0),
            v=[1] * 8, mode='valid') / nvalid)
        arrmda8 = np.ma.masked_invalid(daymax(arra8_m, axis=axis))
    else:
        arra8 = np.apply_along_axis(
            np.convolve, axis=axis, arr=arr, v=[1 / 8.] * 8, mode='valid')
        arrmda8 = daymax(arra8, axis=axis)
    return arrmda8


def _dayfunc(func, arr, axis=None, keepdims=True):
    """
    Arguments:
       func - function with reduceat property
       arr - array_like
       axis - axis overwhich to apply func.reduceat
       keepdims - must be true
    Returns:
       out - array_like with func.reduceat applied at 24 element increments
    """
    daystartend = np.arange(0, arr.shape[axis], 24)
    sliceme = (slice(None),) * axis + (slice(-1),)
    return func.reduceat(arr, daystartend, axis=axis)[sliceme]


def daymax(arr, axis=None, keepdims=True):
    """
    see _dayfunc with np.maximum as func
    """
    return _dayfunc(np.maximum, arr, axis=axis, keepdims=keepdims)


def daymin(arr, axis=None, keepdims=True):
    """
    see _dayfunc with np.minimum as func
    """
    return _dayfunc(np.minimum, arr, axis=axis, keepdims=keepdims)


def daymean(arr, axis=None, keepdims=True):
    """
    see _dayfunc with np.mean as func
    """
    return _dayfunc(np.mean, arr, axis=axis, keepdims=keepdims)


def daystd(arr, axis=None, keepdims=True):
    """
    see _dayfunc with np.std as func
    """
    return _dayfunc(np.std, arr, axis=axis, keepdims=keepdims)


def dayvar(arr, axis=None, keepdims=True):
    """
    see _dayfunc with np.var as func
    """
    return _dayfunc(np.var, arr, axis=axis, keepdims=keepdims)



#!/usr/bin/env python
def main():
    from PseudoNetCDF.pncparse import getparser, pncparse
    from PseudoNetCDF.plotutil.pncmap import makemaps
    parser = getparser(has_ofile=True, map_options=True,
                       plot_options=True, interactive=True)
    ifiles, args = pncparse(
        has_ofile=True, plot_options=True, interactive=True, parser=parser)
    args.ifiles = ifiles
    makemaps(args)


if __name__ == '__main__':
    main()



from __future__ import print_function
from PseudoNetCDF.netcdf import NetCDFFile
from PseudoNetCDF.sci_var import PseudoNetCDFFile
from PseudoNetCDF.MetaNetCDF import file_master
from warnings import warn


def mrgidx(ipr_paths, irr_paths, idx):
    if isinstance(irr_paths, str):
        irrf = NetCDFFile(irr_paths)
    else:
        irrf = file_master([NetCDFFile(irr_path) for irr_path in irr_paths])

    if isinstance(ipr_paths, str):
        iprf = NetCDFFile(ipr_paths)
    else:
        iprf = file_master([NetCDFFile(ipr_path) for ipr_path in ipr_paths])

    # Process and Reaction keys should exclude TFLAG
    pr_keys = [pr for pr in iprf.variables.keys() if pr not in ('TFLAG',)]
    rr_keys = [rr for rr in irrf.variables.keys() if rr not in ('TFLAG',)]

    # Attempt to order reactions by number
    # this is not necessary, but is nice and clean
    try:
        rr_keys = [(int(rr.split('_')[1]), rr) for rr in rr_keys]
        rr_keys.sort()
        rr_keys = [rr[1] for rr in rr_keys]
    except Exception:
        warn("Cannot sort reaction keys")

    # Processes are predicated by a delimiter
    prcs = list(set(['_'.join(pr.split('_')[:-1]) for pr in pr_keys]))
    # Species are preceded by a delimiter
    spcs = list(set(['_'.join(pr.split('_')[-1:]) for pr in pr_keys]))

    # Select a dummy variable for extracting properties
    pr_tmp = iprf.variables[pr_keys[0]]

    # Create an empty file and decorate
    # it as necessary
    outf = PseudoNetCDFFile()
    outf.Species = "".join([spc.ljust(16) for spc in spcs])
    outf.Process = "".join([prc.ljust(16) for prc in prcs])
    outf.Reactions = "".join([rr_key.ljust(16) for rr_key in rr_keys])
    outf.createDimension("PROCESS", len(prcs))
    outf.createDimension("SPECIES", len(spcs))
    outf.createDimension("RXN", len(rr_keys))
    outf.createDimension("TSTEP", pr_tmp[:, 0, 0, 0].shape[0])
    outf.createDimension("TSTEP_STAG", len(outf.dimensions["TSTEP"]) + 1)
    outf.createDimension("ROW", 1)
    outf.createDimension("LAY", 1)
    outf.createDimension("COL", 1)
    outf.createDimension("VAR", 3)
    outf.createDimension("DATE-TIME", 2)
    tflag = outf.createVariable("TFLAG", "i", ('TSTEP', 'VAR', 'DATE-TIME'))
    tflag.__dict__.update(dict(units="<YYYYJJJ,HHDDMM>",
                               var_desc='TFLAG'.ljust(16),
                               long_name='TFLAG'.ljust(16)))
    tflag[:, :, :] = iprf.variables['TFLAG'][:][:, [0], :]
    shape = outf.createVariable("SHAPE", "i", ("TSTEP", "LAY", "ROW", "COL"))
    shape.__dict__.update(dict(units="ON/OFF", var_desc="SHAPE".ljust(16),
                               long_name="SHAPE".ljust(16)))
    shape[:] = 1
    irr = outf.createVariable("IRR", "f", ("TSTEP", "RXN"))
    irr.__dict__.update(dict(units=pr_tmp.units, var_desc="IRR".ljust(16),
                             long_name="IRR".ljust(16)))
    ipr = outf.createVariable("IPR", "f", ("TSTEP", "SPECIES", "PROCESS"))
    irr.__dict__.update(dict(units=pr_tmp.units, var_desc="IPR".ljust(16),
                             long_name="IPR".ljust(16)))

    for rr, var in zip(rr_keys, irr.swapaxes(0, 1)):
        var[:] = irrf.variables[rr][:][idx]

    for prc, prcvar in zip(prcs, ipr.swapaxes(0, 2)):
        for spc, spcvar in zip(spcs, prcvar):
            try:
                spcvar[:] = iprf.variables['_'.join([prc, spc])][:][idx]
            except KeyError as es:
                warn(str(es))

    return outf


if __name__ == '__main__':
    from PseudoNetCDF.pncdump import pncdump
    import sys

    x = ', '.join(sys.argv[1:])
    try:
        pncdump(eval('mrgidx(%s)' % (x,)), name=mrgidx)
    except Exception:
        print("\nUsage: python -m PseudoNetCDF.cmaqfiles.pa \"ipr_paths, " +
              "irr_paths, idx\"\n  ipr_paths - a single string or list of " +
              "strings\n  irr_paths - a single string or list of strings\n " +
              " idx - 4D numpy slice indexes a time series\n\n\nExample:\n " +
              " $ python -m PseudoNetCDF.cmaqfiles.pa \"['CMAQ_IPR1.nc'," +
              "'CMAQ_IPR2.nc'], 'CMAQ_IRR.nc', (slice(None),0,1,1)\"",
              file=sys.stderr)



import numpy as np
_coorddict = dict(west_east='longitude', south_north='latitude', Time='time',
                  bottom_top='altitude', west_east_stag='longitude',
                  south_north_stag='latitude', Time_stag='time',
                  bottom_top_stag='altitude',)


def add_cf_from_wrfioapi(ifile, coordkeys=[]):
    try:
        for invark, outvark in [('XLONG', 'longitude'), ('XLAT', 'latitude')]:
            try:
                invar = ifile.variables[invark]
            except KeyError:
                invark += '_M'
                invar = ifile.variables[invark]
            outvar = ifile.createVariable(
                outvark, invar.dtype.char, invar.dimensions[1:])
            for pk in invar.ncattrs():
                setattr(outvar, pk, getattr(invar, pk))
            outvar[:] = invar[0]
    except Exception as e:
        print(e)
        pass

    # Add time coordinate if not available.
    try:
        outvar = ifile.createVariable('time', 'i', ('Time',))
        invar = ifile.variables['Times']
        for pk in invar.ncattrs():
            setattr(outvar, pk, getattr(invar, pk))
        outvar.units = 'seconds since 1985-01-01 00:00:00Z'
        invals = invar[:].copy().view('S19')
        invals = np.char.decode(invals)
        sdate = np.datetime64('1985-01-01T00:00:00Z')
        timesince = (np.array([np.datetime64(time[0].replace(
            '_', 'T') + 'Z') for time in invals]) - sdate).astype('l')
        outvar[:] = timesince
    except Exception as e:
        print(e)
        pass
    # add projection
    grid_mapping_name = get_proj(ifile)
    # Add x and y coordinates
    try:
        gridvar = ifile.variables[grid_mapping_name]
        xvar = ifile.createVariable('x', 'd', ('west_east',))
        xvar.units = 'm'
        xvar.standard_name = 'x'
        xvar[:] = np.arange(xvar.size, dtype='d') * ifile.DX - \
            gridvar.false_easting + ifile.DX / 2
        yvar = ifile.createVariable('y', 'd', ('south_north',))
        yvar.units = 'm'
        yvar.standard_name = 'y'
        yvar[:] = np.arange(yvar.size, dtype='d') * ifile.DY - \
            gridvar.false_northing + ifile.DY / 2
        wevar = ifile.createVariable('west_east', 'd', ('west_east',))
        wevar.units = 'm'
        wevar.standard_name = 'west_east'
        wevar[:] = xvar[:] + gridvar.false_easting
        snvar = ifile.createVariable('south_north', 'd', ('south_north',))
        snvar.units = 'm'
        snvar.standard_name = 'south_north'
        snvar[:] = yvar[:] + gridvar.false_northing
    except Exception as e:
        raise e
    for k in ifile.variables.keys():
        var = ifile.variables[k]
        olddims = [dk for dk in var.dimensions]
        newdims = [_coorddict.get(dk, dk) for dk in var.dimensions]
        if olddims != newdims:
            var.coordinates = ' '.join(newdims)
            var.grid_mapping = grid_mapping_name

    ifile.Conventions = 'CF-1.6'


def get_proj(ifile):
    """
    MAP_PROJ - Model projection [1=Lambert, 2=polar stereographic,
                                 3=mercator, 6=lat-lon]  (required)
    TRUELAT1 - required for MAP_PROJ = 1, 2, 3 (defaults to 0 otherwise)
    TRUELAT2 - required for MAP_PROJ = 6 (defaults to 0 otherwise)
    STAND_LON - Standard longitude used in model projection (required)
    REF_LON, REF_LON - A reference longitude and latitude (required)
    KNOWNI, KNOWNJ - The I and J locations of REF_LON and REF_LAT (required)
    POLE_LAT - optional for MAP_PROJ = 6 (defaults to 90 otherwise)
    POLE_LAT - optional for MAP_PROJ = 6 (defaults to 0 otherwise)
    DX, DY - required for MAP_PROJ = 1, 2, 3 (defaults to 0 otherwise)
    LATINC, LONINC - required for MAP_PROJ = 6 (defaults to 0 otherwise)
    """

    projname = {1: "lambert_conformal_conic",
                2: 'polar_stereographic', 3: 'mercator'}[ifile.MAP_PROJ]
    var = ifile.createVariable(projname, 'i', ())
    var.grid_mapping_name = projname
    var.earth_radius = 6370000.
    var.false_easting = 0
    var.false_northing = 0
    x0_from_cen = len(ifile.dimensions['west_east']) / 2 * ifile.DX
    y0_from_cen = len(ifile.dimensions['south_north']) / 2 * ifile.DY
    if projname == 'mercator':
        var.longitude_of_projection_origin = ifile.STAND_LON
        var.standard_parallel = ifile.TRUELAT1
    elif projname == "lambert_conformal_conic":
        var.standard_parallel = np.array([ifile.TRUELAT1, ifile.TRUELAT2])
        var.longitude_of_central_meridian = ifile.STAND_LON
        var.latitude_of_projection_origin = ifile.MOAD_CEN_LAT
    elif projname == 'polar_stereographic':
        var.straight_vertical_longitude_from_pole = ifile.STAND_LON
        var.latitude_of_projection_origin = ifile.MOAD_CEN_LAT
        var.standard_parallel = ifile.TRUELAT1

    from PseudoNetCDF.coordutil import getproj4_from_cf_var
    from mpl_toolkits.basemap import pyproj
    projstr = getproj4_from_cf_var(var)
    proj = pyproj.Proj(projstr)
    gcx, gcy = proj(ifile.CEN_LON, ifile.CEN_LAT)
    glx = gcx - x0_from_cen
    gly = gcy - y0_from_cen
    var.false_easting = -glx
    var.false_northing = -gly
    return projname



__all__ = ['lateral_boundary']
__doc__ = """
.. _Memmap
:mod:`Memmap` -- lateral_boundary Memmap interface
============================================

.. module:: Memmap
   :platform: Unix, Windows
   :synopsis: Provides :ref:`PseudoNetCDF` memory map for CAMx
              lateral_boundary files.  See PseudoNetCDFFile
              for interface details
.. moduleauthor:: Barron Henderson <barronh@unc.edu>
"""

# Distribution packages
import unittest
import struct

# Site-Packages
from numpy import zeros, array, memmap, dtype, testing

# This Package modules
from PseudoNetCDF.sci_var import PseudoNetCDFFile, PseudoIOAPIVariable
from PseudoNetCDF.sci_var import PseudoNetCDFVariables
from PseudoNetCDF.ArrayTransforms import ConvertCAMxTime
from PseudoNetCDF.camxfiles.units import get_uamiv_units

# for use in identifying uncaught nan
listnan = struct.unpack('>f', b'\xff\xc0\x00\x00')[0]
checkarray = zeros((1,), 'f')
checkarray[0] = listnan
array_nan = checkarray[0]


class lateral_boundary(PseudoNetCDFFile):
    """
    lateral_boundary provides a PseudoNetCDF interface for CAMx
    lateral_boundary files.  Where possible, the inteface follows
    IOAPI conventions (see www.baronams.com).

    ex:
        >>> lateral_boundary_path = 'camx_lateral_boundary.bin'
        >>> rows,cols = 65,83
        >>> lbf = lateral_boundary(lateral_boundary_path,rows,cols)
        >>> lbf.variables.keys()
        ['TFLAG', 'O3', 'NO', 'NO2', ...]
        >>> tflag = lbf.variables['TFLAG']
        >>> tflag.dimensions
        ('TSTEP', 'VAR', 'DATE-TIME')
        >>> tflag[0,0,:]
        array([2005185,       0])
        >>> tflag[-1,0,:]
        array([2005185,  240000])
        >>> v = lbf.variables['O3']
        >>> v.dimensions
        ('TSTEP', 'LAY', 'ROW', 'COL')
        >>> v.shape
        (25, 28, 65, 83)
        >>> lbf.dimensions
        {'TSTEP': 25, 'LAY': 28, 'ROW': 65, 'COL': 83}
    """

    __emiss_hdr_fmt = dtype(dict(names=['SPAD', 'name', 'note', 'itzon',
                                        'nspec', 'ibdate', 'btime', 'iedate',
                                        'etime', 'EPAD'],
                                 formats=['>i', '(10,4)>S1', '(60,4)>S1', '>i',
                                          '>i', '>i', '>f', '>i', '>f', '>i']))
    __grid_hdr_fmt = dtype(dict(names=['SPAD', 'plon', 'plat', 'iutm', 'xorg',
                                       'yorg', 'delx', 'dely', 'nx', 'ny',
                                       'nz', 'iproj', 'istag', 'tlat1',
                                       'tlat2', 'rdum5', 'EPAD'],
                                formats=['>i', '>f', '>f', '>i', '>f', '>f',
                                         '>f', '>f', '>i', '>i', '>i', '>i',
                                         '>i', '>f', '>f', '>f', '>i']))
    __cell_hdr_fmt = dtype(dict(names=['SPAD', 'ione1', 'ione2', 'nx', 'ny',
                                       'EPAD'],
                                formats=['>i', '>i', '>i', '>i', '>i', '>i']))
    __time_hdr_fmt = dtype(dict(names=['SPAD', 'ibdate', 'btime', 'iedate',
                                       'etime', 'EPAD'],
                                formats=['>i', '>i', '>f', '>i', '>f', '>i']))
    __spc_fmt = dtype("(10,4)>S1")

    __ione = 1
    __idum = 0
    __rdum = 0.

    @classmethod
    def isMine(cls, path):
        try:
            # initialization calls readheader, which has an assertion check
            lateral_boundary(path)
            return True
        except Exception:
            return False

    def __init__(self, rf, mode='r'):
        """
        Initialization included reading the header and learning
        about the format.

        see __readheader and __gettimestep() for more info
        """
        self.__rffile = rf
        self.__mode = mode

        self.createDimension('DATE-TIME', 2)

        self.__readheader()

        # Add IOAPI metavariables
        self.NLAYS = len(self.dimensions['LAY'])
        self.NROWS = len(self.dimensions['ROW'])
        self.NCOLS = len(self.dimensions['COL'])
        self.NVARS = len(self.dimensions['VAR'])
        self.NSTEPS = len(self.dimensions['TSTEP'])
        varlist = "".join([i.ljust(16) for i in self.__var_names__])
        setattr(self, 'VAR-LIST', varlist)
        self.GDTYP = 2
        name = self.__emiss_hdr['name'][0, :, 0].copy().view('S10')[0]
        note = self.__emiss_hdr['note'][0, :, 0].copy().view('S60')[0]
        if hasattr(name, 'decode'):
            name = name.decode()
        if hasattr(note, 'decode'):
            note = note.decode()
        self.NAME = name
        self.NOTE = note
        self.ITZON = self.__emiss_hdr['itzon'][0]

        # Create variables
        self.variables = PseudoNetCDFVariables(
            self.__variables, self.__var_names__ + ['TFLAG', 'ETFLAG'])
        self.variables['TFLAG'] = ConvertCAMxTime(
            self.__memmap__['DATE']['BDATE'],
            self.__memmap__['DATE']['BTIME'], self.NVARS)
        self.variables['ETFLAG'] = ConvertCAMxTime(
            self.__memmap__['DATE']['EDATE'],
            self.__memmap__['DATE']['BTIME'], self.NVARS)

        self.SDATE, self.STIME = self.variables['TFLAG'][0, 0, :]

    def __checkfilelen(self):
        f = open(self.__rffile, 'rb')
        f.seek(0, 2)
        flen = f.tell()
        f.close()
        return flen

    def __readheader(self):
        offset = 0
        self.__emiss_hdr = memmap(
            self.__rffile, mode=self.__mode, dtype=self.__emiss_hdr_fmt,
            shape=1, offset=offset)
        nspec = self.__emiss_hdr['nspec'][0]
        offset += self.__emiss_hdr.dtype.itemsize * self.__emiss_hdr.size

        self.__grid_hdr = memmap(
            self.__rffile, mode=self.__mode, dtype=self.__grid_hdr_fmt,
            shape=1, offset=offset)

        self.XORIG = self.__grid_hdr['xorg'][0]
        self.YORIG = self.__grid_hdr['yorg'][0]
        self.XCELL = self.__grid_hdr['delx'][0]
        self.YCELL = self.__grid_hdr['dely'][0]
        self.PLON = self.__grid_hdr['plon'][0]
        self.PLAT = self.__grid_hdr['plat'][0]
        self.TLAT1 = self.__grid_hdr['tlat1'][0]
        self.TLAT2 = self.__grid_hdr['tlat2'][0]
        self.IUTM = self.__grid_hdr['iutm'][0]
        self.ISTAG = self.__grid_hdr['istag'][0]
        self.CPROJ = self.__grid_hdr['iproj'][0]

        # Map CAMx projection constants to IOAPI
        GDTYPE = self.GDTYP = {0: 1, 1: 5, 2: 2, 3: 6}[
            self.__grid_hdr['iproj'][0]]
        self.XCENT = self.__grid_hdr['plon'][0]
        self.YCENT = self.__grid_hdr['plat'][0]
        if GDTYPE in (1, 2):
            self.P_ALP = self.__grid_hdr['tlat1'][0]
            self.P_BET = self.__grid_hdr['tlat2'][0]
            self.P_GAM = self.__grid_hdr['plon'][0]
        elif GDTYPE == 5:
            self.P_ALP = self.__grid_hdr['iutm'][0]
            self.P_BET = 0.
            self.P_GAM = 0.
        elif GDTYPE == 6:
            self.P_ALP = {90: 1, -90: -1}[self.__grid_hdr['plat'][0]]
            self.P_BET = self.__grid_hdr['tlat1'][0]
            self.P_GAM = self.__grid_hdr['plon'][0]
        else:
            raise ValueError('Unknown projection')

        nx = self.__grid_hdr['nx'][0]
        ny = self.__grid_hdr['ny'][0]
        nz = max(self.__grid_hdr['nz'], array([1]))[0]

        offset += self.__grid_hdr.dtype.itemsize * self.__grid_hdr.size
        self.__cell_hdr = memmap(
            self.__rffile, mode=self.__mode, dtype=self.__cell_hdr_fmt,
            shape=1, offset=offset)

        offset += self.__cell_hdr.dtype.itemsize * self.__cell_hdr.size + 4
        self.__spc_hdr = memmap(self.__rffile, mode=self.__mode,
                                dtype=self.__spc_fmt, shape=nspec,
                                offset=offset)

        offset += self.__spc_hdr.dtype.itemsize * self.__spc_hdr.size + 4
        self._boundary_def = {}
        for bkey, bdim in [('WEST', ny), ('EAST', ny), ('SOUTH', nx),
                           ('NORTH', nx)]:
            __bound_fmt = dtype(dict(names=['SPAD', 'ione', 'iedge', 'ncell',
                                            'edgedata', 'EPAD'],
                                     formats=['>i', '>i', '>i', '>i',
                                              '(%d,%d)>i' % (bdim, 4), '>i']))
            self._boundary_def[bkey] = memmap(
                self.__rffile, mode=self.__mode, dtype=__bound_fmt, shape=1,
                offset=offset)
            assert(self._boundary_def[bkey]['SPAD'] ==
                   (__bound_fmt.itemsize - 8))
            offset += __bound_fmt.itemsize

        date_time_fmt = dtype(dict(names=['SPAD', 'BDATE', 'BTIME', 'EDATE',
                                          'ETIME', 'EPAD'],
                                   formats=['>i', '>i', '>f', '>i', '>f',
                                            '>i']))
        date_time_block_size = 6
        spc_we_fmt = dtype(dict(names=['SPAD', 'IONE', 'SPC', 'IEDGE', 'DATA',
                                       'EPAD'],
                                formats=['>i', '>i', '(10,4)>S1', '>i',
                                         '(%d,%d)>f' % (ny, nz), '>i']))
        spc_sn_fmt = dtype(dict(names=['SPAD', 'IONE', 'SPC', 'IEDGE', 'DATA',
                                       'EPAD'],
                                formats=['>i', '>i', '(10,4)>S1', '>i',
                                         '(%d,%d)>f' % (nx, nz), '>i']))

        spc_lat_fmt = dtype(dict(names=['WEST', 'EAST', 'SOUTH', 'NORTH'],
                                 formats=[spc_we_fmt, spc_we_fmt, spc_sn_fmt,
                                          spc_sn_fmt, ]))
        # Get species names from spc_hdr
        self.__var_names__ = []
        spc_names = [spc[:, 0].copy().view('S10')[0] for spc in self.__spc_hdr]
        spc_names = [spc.decode() if hasattr(spc, 'decode')
                     else spc for spc in spc_names]
        self.__spc_names__ = [spc.strip() for spc in spc_names]
        for spc in self.__spc_names__:
            for bkey in ['WEST_', 'EAST_', 'SOUTH_', 'NORTH_']:
                self.__var_names__.append(bkey + spc)

        spc_lat_block_size = spc_lat_fmt.itemsize // 4
        data_block_fmt = dtype(dict(
            names=['DATE'] + self.__spc_names__,
            formats=[date_time_fmt] + [spc_lat_fmt] * nspec))

        data_block_size = date_time_block_size + nspec * spc_lat_block_size
        f = open(self.__rffile)
        f.seek(0, 2)
        size = f.tell()
        f.close()
        del f
        ntimes = float(size - offset) // 4. // data_block_size
        if int(ntimes) != ntimes:
            raise ValueError("Not an even number of times %f" % ntimes)
        ntimes = int(ntimes)
        self.createDimension('LAY', nz)
        self.createDimension('COL', nx)
        self.createDimension('ROW', ny)
        self.createDimension('TSTEP', ntimes)
        self.createDimension('VAR', nspec * 4)

        self.__memmap__ = memmap(
            self.__rffile, mode=self.__mode, dtype=data_block_fmt,
            offset=offset)

    def __variables(self, k):
        edgename = k.split('_')[0]
        spcname = k[len(edgename) + 1:]
        if edgename in ('WEST', 'EAST'):
            dimensions = ('TSTEP', 'ROW', 'LAY')
        else:
            dimensions = ('TSTEP', 'COL', 'LAY')
        outvals = self.__memmap__[spcname][edgename]['DATA']
        units = get_uamiv_units(self.NAME, spcname)

        return PseudoIOAPIVariable(self, k, 'f', dimensions, values=outvals,
                                   units=units)


class TestMemmap(unittest.TestCase):
    def runTest(self):
        pass

    def setUp(self):
        pass

    def testLB(self):
        import PseudoNetCDF.testcase
        latfile = lateral_boundary(
            PseudoNetCDF.testcase.camxfiles_paths['lateral_boundary'])
        latfile.variables['TFLAG']
        aassert = testing.assert_array_almost_equal
        checkw = array([5.09086549e-02, 5.15854955e-02, 5.20327762e-02,
                        4.51020785e-02, 4.81765866e-02, 5.01902141e-02,
                        5.11084236e-02, 5.30455858e-02, 5.39216697e-02,
                        5.49603552e-02, 5.52441478e-02, 5.53695373e-02,
                        4.70436066e-02, 4.88652885e-02, 5.00564687e-02,
                        4.30268869e-02, 4.65402082e-02, 4.93281633e-02,
                        4.84376177e-02, 5.15173525e-02, 5.32316864e-02,
                        5.27723655e-02, 5.43252379e-02, 5.51613718e-02],
                       dtype='f').reshape(2, 4, 3)
        aassert(latfile.variables['WEST_O3'], checkw)
        checke = array([4.32289392e-02, 4.32576202e-02, 4.32824045e-02,
                        4.28533703e-02, 4.29217815e-02, 4.29747701e-02,
                        4.20787595e-02, 4.21368703e-02, 4.21847440e-02,
                        4.20016758e-02, 4.20646742e-02, 4.21276242e-02,
                        4.45125513e-02, 4.45423163e-02, 4.45689932e-02,
                        4.34477255e-02, 4.35189568e-02, 4.35763896e-02,
                        4.23655175e-02, 4.24236283e-02, 4.24756072e-02,
                        4.21473905e-02, 4.22183946e-02, 4.22968678e-02],
                       dtype='f').reshape(2, 4, 3)
        aassert(latfile.variables['EAST_O3'], checke)
        checks = array([5.00756986e-02, 5.07836118e-02, 5.12398519e-02,
                        4.53975834e-02, 4.85784635e-02, 5.06106876e-02,
                        4.88935970e-02, 5.18746264e-02, 5.34388497e-02,
                        5.52802160e-02, 5.66427484e-02, 5.73641062e-02,
                        5.76447174e-02, 5.91668785e-02, 5.98937273e-02,
                        4.64089997e-02, 4.85925563e-02, 4.98120859e-02,
                        4.31438088e-02, 4.85876575e-02, 5.05158082e-02,
                        4.63983566e-02, 5.09891734e-02, 5.34970984e-02,
                        5.16217351e-02, 5.49310222e-02, 5.70215993e-02,
                        5.37858233e-02, 5.71687669e-02, 5.90757653e-02],
                       dtype='f').reshape(2, 5, 3)
        aassert(latfile.variables['SOUTH_O3'], checks)
        checkw = array([4.67120931e-02, 4.72769439e-02, 4.75757420e-02,
                        4.63561714e-02, 4.69456315e-02, 4.72794324e-02,
                        4.45141681e-02, 4.53266129e-02, 4.58004810e-02,
                        4.20911871e-02, 4.29873653e-02, 4.35210876e-02,
                        3.29810269e-02, 3.38455141e-02, 3.43711227e-02,
                        4.65101935e-02, 4.71580848e-02, 4.75305654e-02,
                        4.55248095e-02, 4.63739596e-02, 4.68796641e-02,
                        4.44590077e-02, 4.58386838e-02, 4.66508269e-02,
                        4.33906466e-02, 4.47074845e-02, 4.51488644e-02,
                        3.23279053e-02, 3.31198536e-02, 3.36514898e-02],
                       dtype='f').reshape(2, 5, 3)
        aassert(latfile.variables['NORTH_O3'], checkw)

    def testNCF2LB(self):
        import PseudoNetCDF.testcase
        from PseudoNetCDF.pncgen import pncgen
        import os
        inpath = PseudoNetCDF.testcase.camxfiles_paths['lateral_boundary']
        outpath = inpath + '.check'
        infile = lateral_boundary(inpath)
        pncgen(infile, outpath, format='camxfiles.lateral_boundary')
        orig = open(inpath, 'rb').read()
        new = open(outpath, 'rb').read()
        assert(orig == new)
        os.remove(outpath)


if __name__ == '__main__':
    lb = lateral_boundary('../../testcase/utils/CAMx/lateral_boundary/' +
                          'BC.vistas_2002gt2a_STL_36_68X68_16L.2002154')
    # unittest.main()



from __future__ import print_function
import numpy as np
from PseudoNetCDF import PseudoNetCDFFile, PseudoNetCDFVariable
from PseudoNetCDF import PseudoNetCDFVariables
from PseudoNetCDF.coordutil import gettimes
from PseudoNetCDF._getwriter import registerwriter
from datetime import datetime
from collections import OrderedDict
from warnings import warn
dtype = np.dtype


def _vord(val):
    try:
        vnew = ord(val)
    except TypeError:
        if len(val) == 0:
            vnew = 0.
        else:
            raise
    return vnew


_vordv = np.vectorize(_vord)


def timeit(key):
    from time import time
    if key in timeit.starts:
        started = timeit.starts.pop(key)
        print(key, (time() - started) / 60., 'min')
    else:
        print(key, time(), 'seconds since epoch')
        timeit.starts[key] = time()


timeit.starts = {}

thdtype = dtype([('YYMMDDHHFF', '>10S'), ('LEVEL', '>2S'), ('GRID', '>2S'),
                 ('INDX', '>4S'), ('Z1', '>4S'), ('MB1', '>14S'),
                 ('MB2', '>14S'), ('SRCE', '>S4'), ('MGRID', '>S3'),
                 ('VSYS', '>S2'), ('POLLAT', '>S7'), ('POLLON', '>S7'),
                 ('REFLAT', '>S7'), ('REFLON', '>S7'),
                 ('GRIDX', '>S7'), ('ORIENT', '>S7'), ('TANLAT', '>S7'),
                 ('SYNCHX', '>S7'), ('SYNCHY', '>S7'), ('SYNCHLAT', '>S7'),
                 ('SYNCHLON', '>S7'), ('RESERVED', '>S7'), ('NX', '>S3'),
                 ('NY', '>S3'), ('NZ', '>S3'), ('VSYS2', '>S2'),
                 ('LENH', '>S4')])
vhdtype = dtype([('YYMMDDHHFF', '>10S'), ('LEVEL', '>2S'), ('grid', '>2S'),
                 ('VKEY', '>4S'), ('EXP', '>4S'), ('PREC', '>14S'),
                 ('VAR1', '>14S')])

stdprops = dict(
    x=('x', 'm'),
    y=('y', 'm'),
    longitude=('longitude', 'degrees'),
    latitude=('latitude', 'degrees'),
    longitude_bounds=('longitude_bounds', 'degrees'),
    latitude_bounds=('latitude_bounds', 'degrees'),
    ABSV=('ABSOLUTE VORTICITY', '100000.00 ? /s'),
    CAPE=('CONVECTIVE AVAILABLE POTENTIAL ENERGY', 'J/kg'),
    CFZR=('CATEGORIAL FREEZING RAIN (YES=1/NO=0)', '0--1'),
    CICE=('CATEGORIAL ICE PELLETS (YES=1/NO=0)', '0--1'),
    CINH=('CONVECTIVE INHIBITION', 'J/kg'),
    CLDB=('CLOUD BOTTOM HEIGHT', 'm'),
    CLDT=('CLOUD TOP HEIGHT', 'm'),
    CONC=('CONCENTRATION', '/m3'),
    CPP3=('3-HOUR ACC. CONVECTIVE PRECIPITATION', 'mm'),
    CPP6=('6-HOUR ACC. CONVECTIVE PRECIPITATION', 'mm'),
    CPPT=('CONVECTIVE PRECIPITATION', 'mm'),
    CRAI=('CATEGORIAL RAIN (YES=1/NO=0)', '0--1'),
    CSNO=('CATEGORIAL SNOW (YES=1/NO=0)', '0--1'),
    CWMR=('CLOUD WATER MIXING RATIO', 'g/kg'),
    DEPO=('SURFACE DEPOSITION', '/m2'),
    DEPS=('SURFACE DEPOSITION', '/m2'),
    DEPV=('DEPOSITION VELOCITY', 'm/s'),
    DLWF=('DOWNWARD LONG WAVE RADIATION FLUX', 'W/m2'),
    DP2M=('2 M DEW POINT', 'degC'),
    DSWF=('DOWNWARD SHORT WAVE RADIATION FLUX', 'W/m2'),
    EXCO=('EXCHANGE COEFFICIENT', '1000.00 ? GM2S'),
    FLAG=('WIND FLAGS', 'KNTS'),
    HCLD=('HIGH CLOUD COVER', 'PCT'),
    HFLX=('LATENT HEAT FLUX', 'W/m2'),
    HGT1=('1000 MB HEIGHT', 'm'),
    HGT5=('500 MB HEIGHT', 'm'),
    HGTS=('Geopotential height', 'gpm*'),
    ICWT=('ICE-COVERED WATER', '0--1'),
    LCLD=('LOW CLOUD COVER', 'PCT'),
    LHTF=('LATENT HEAT NET FLUX', 'W/m2'),
    LIB4=('BEST 4-LAYER LIFTED INDEX', 'K'),
    LISD=('STANDARD LIFTED INDEX', 'degC'),
    LTHF=('Latent heat flux', 'W/m2'),
    LTNG=('LIGHTNING STRIKES', '#'),
    MCLD=('MEDIUM CLOUD COVER', 'PCT'),
    MOMF=('TOTAL MOMENTUM FLUX', 'N/m2'),
    MSLP=('MEAN SEA-LEVEL PRESSURE', 'hPa'),
    MXHT=('MIXED LAYER DEPTH', 'm'),
    MXLR=('NUMBER OF MIXED SIGMA LAYERS', '0--N'),
    OPPT=('OBSERVED PRECIPITATION', 'mm'),
    P10M=('POTENTIAL TEMPERATURE ', 'K'),
    PBLH=('PLANETARY BOUNDARY LAYER HEIGHT', 'm'),
    PRAT=('PRECIPITATION RATE', 'KGM2'),
    PRES=('PRESSURE', 'hPa'),
    PRSS=('SURFACE PRESSURE', 'hPa'),
    PTYP=('PRECIPITATION TYPE (RA=1,TRW=2,ZR=3,ICE=4,SN=5)', '1--5'),
    QSTR=('FLUX MIXING RATIO', '1000.00 ? G/KG'),
    REFC=('COMPOSITE REFLECTIVITY', 'dBZ'),
    RELH=('RELATIVE HUMIDITY', 'PCT'),
    RGHS=('SURFACE ROUGHNESS', 'm'),
    RH2M=('2 METER RELATIVE HUMIDITY', 'PCT'),
    SFCC=('SURFACE CONCENTRATION', '/m3'),
    SHGT=('SURFACE HEIGHT', 'm'),
    SHTF=('SENSIBLE HEAT NET FLUX', 'W/m2'),
    SNOC=('SNOW COVERAGE ', 'PCT'),
    SNOW=('SNOW COVERAGE', '0--1'),
    SNWD=('SNOW DEPTH', 'CM'),
    SOLM=('SOIL MOISTURE', 'kg/m2'),
    SOLT=('SOIL TEMPERATURE ', 'degC'),
    SOLW=('0 TO 200 CM SOIL MOISTURE CONTENT', 'kg/m2'),
    SPHU=('SPECIFIC HUMIDITY', 'kg/kg'),
    STRM=('STREAMLINES ', 'KNTS'),
    T02M=('Temperature at 2 m', 'K'),
    TCLD=('AVERAGE TOTAL CLOUD COVER', 'PCT'),
    TEMP=('Temperature', 'K'),
    THKN=('THICKNESS', '0.10 ? dm'),
    THTS=('SURFACE POTENTIAL TEMPERATURE', 'DEFC'),
    TKEN=('TOTAL KINETIC ENERGY', 'J'),
    TMPS=('Temperature at surface', 'K'),
    TOPO=('TOPOGRAPHY', 'm'),
    TPP1=('1 HOUR ACCUMULATED PRECIPITATION', 'm'),
    TPP3=('3-HOUR ACCUMULATED PRECIPITATION', 'mm'),
    TPP6=('Total precipitation (6-h)', 'm'),
    TPPA=('TOTAL ACCUMULATED PRECIPITATION', 'mm'),
    TPPT=('12-HOUR ACCUMULATED PRECIPITATION', 'mm'),
    TSTR=('FLUX TEMPERATURE', 'degC'),
    U10M=('10 M U-WIND COMPONENT', 'm/s'),
    ULWF=('UPWARD LONG WAVE RADIATION FLUX', 'W/m2'),
    UMOF=('MOMENTUM FLUX, U-WIND COMPONENT', 'N/m2'),
    USTR=('FRICTION VELOCITY', '100.00 ? m/s'),
    UWND=('U-WIND COMPONENT', 'm/s'),
    V10M=('10 M V-WIND COMPONENT', 'm/s'),
    VGTP=('VEGETATION TYPE', '0--1'),
    VMOF=('MOMENTUM FLUX, V-WIND COMPONENT', 'N/m2'),
    VSBY=('VISIBILITY', 'm'),
    VWND=('V-WIND COMPONENT', 'm/s'),
    WESD=('WATER EQUIV. OF ACCUM. SNOW DEPTH', 'kg/m2'),
    WFLX=('WATER VAPOR FLUX', 'kg m2 s'),
    WSPD=('WIND SPEED', 'KNTS'),
    WTMP=('WATER TEMPERATURE', 'degC'),
    WVCT=('WIND VECTORS', 'KNTS'),
    WWND=('W-WIND COMPONENT', 'hPa/s'),
)
# not used in favor of defn on S141.htm
#    HGTS = ('HEIGHT ', '0.10 ? DM'),
#    TMPS = ('SURFACE TEMPERATURE', 'degC'),
#    T02M = ('2 M TEMPERATURE', 'degC'),
#    TEMP = ('TEMPERATURE', 'degC'),
#    TMPS = ('SURFACE TEMPERATURE', 'degC'),
#    TPP6 = ('6-HOUR ACCUMULATED PRECIPITATION', 'mm'),

# not used in favor of defn on https://www.ready.noaa.gov/READYflddesc.php
#    SPHU = ('Specific humidity', 'unknown'),


def readvardef(vheader, out={}):
    out['keys'] = keys = OrderedDict()
    out['checksums'] = checksums = OrderedDict()
    out['vglvls'] = vglvls = []
    while not np.char.replace(vheader, b' ', b'') == b'':
        vglvl = float(vheader[:6])
        vgvars = int(vheader[6:8])
        vheader = vheader[8:]
        vglvls.append(vglvl)
        keys[vglvl] = []
        for vgvari in range(vgvars):
            vgkey = vheader[:4]
            checksum = int(vheader[4:7])
            keys[vglvl].append(vgkey)
            checksums[vglvl, vgkey] = checksum
            vheader = vheader[8:]
    out['sfckeys'] = keys[vglvls[0]]
    out['laykeys'] = [(vglvl, keys[vglvl]) for vglvl in vglvls[1:]]
    return out


def writevardef(vglvls, keys, checksums):
    """
    vglvls - iterables of floats [1, .98, ...]
    keys - dictionary of keys by level {vglvli: [vark, ....], ...}
    checksums - integer checsums for each lvl, key --
                {(vglvl, vark): checksum, ...}
    """
    vardef = ''
    vgtxts = getvgtxts(vglvls)

    for vglvl, vgtxt in zip(vglvls, vgtxts):
        vardef += vgtxt[:6]
        lvlvars = keys[vglvl]
        vardef += '%2d' % len(lvlvars)
        for varkey in lvlvars:
            checksum = checksums[vglvl, varkey]
            vardef += varkey.ljust(4)[:4].decode()
            vardef += '%3d' % checksum
            vardef += ' '

    return vardef.ljust(len(vardef) + 108)


def inqarlpackedbit(path):
    f = open(path, 'r')
    fheader = np.fromfile(f, count=1, dtype=thdtype)[0]
    out = {}
    out['LENH'] = hlen = int(fheader['LENH'])
    gridx_off = max(0, (fheader['GRID'][0] - 64) * 1000)
    gridy_off = max(0, (fheader['GRID'][1] - 64) * 1000)
    out['NX'] = int(fheader['NX']) + gridx_off
    out['NY'] = int(fheader['NY']) + gridy_off
    out['NZ'] = int(fheader['NZ'])
    vheader = np.fromfile(f, count=1, dtype='>%dS' % hlen)[0]
    readvardef(vheader, out)
    return out


def getvgtxts(vglvls):
    vgtxts = []
    for vglvl in vglvls:
        if vglvl == 0:
            dp = 0
        else:
            dp = np.floor(np.log10(vglvl) + 1)
        tmpl = '%%6.%df' % np.minimum(5, (5 - dp))
        vgtxt = (tmpl % vglvl)[-6:]
        if len(vgtxt) != 6:
            raise ValueError('Unable to format vglvl:' + str(vglvl))
        vgtxts.append(vgtxt)
    return vgtxts


def writearlpackedbit(infile, path):
    """
    path - path to existing arl packed bit file or location for new file
    infile - NetCDF-like file with
        - vertical 2-D (surface) and 3-D (layers) variables
          with 4 character names
        - a z variable with vertical coordinates
        - all properties from the first index record

    """
    requiredkeys = list(thdtype.names)
    for key in requiredkeys:
        getattr(infile, key)

    svars = {}
    lvars = {}
    props = {}
    sfckeys = props['sfckeys'] = []
    laykeys = props['laykeys'] = []
    for vark, var in infile.variables.items():
        if len(var.shape) == 3:
            svars[vark] = var
            sfckeys.append(vark.encode('ascii'))
        elif len(var.shape) == 4:
            lvars[vark] = var
            lvar = var
            laykeys.append(vark.encode('ascii'))
        else:
            pass

    vglvls = np.append(float(infile.SFCVGLVL),
                       infile.variables['z'][:].array())
    # vgtxts = getvgtxts(vglvls)
    # plus one includes surface
    props['NZ'] = lvar.shape[1] + 1
    props['NY'] = lvar.shape[2]
    props['NX'] = lvar.shape[3]
    datamap = maparlpackedbit(
        path, mode='write', shape=(lvar.shape[0],), props=props)

    theads = datamap['timehead']
    # for key in thdtype.names:
    #    theads[key] = getattr(infile, key)

    # vardefs = datamap['vardef']
    # for ti, vardef in enumerate(vardefs):
    #     sfcvardeftxt = vgtxts[0] + '%2d' % len(svars) +
    #                    ''.join(['%-4s -1 ' % skey.decode()
    #                             for skey in sfckeys])
    #     layvardeftxt = ''
    #     for vgtxt in vgtxts[1:]:
    #         layvardeftxt += vgtxt + '%2d' % len(lvars) +
    #                         ''.join(['%-4s -1 ' % lkey.decode()
    #                                  for lkey in laykeys])
    #
    #
    # vardeftxt = sfcvardeftxt + layvardeftxt
    # defsize = 8+8*len(svars)+(8+8*len(lvars))*len(vgtxts[1:])
    # assert(len(vardeftxt) == defsize)
    # vardefs[:] = vardeftxt
    YYMMDDHHFF = getattr(infile, 'YYMMDDHHFF', '0000000000')
    FF = YYMMDDHHFF[-2:]
    times = gettimes(infile)

    checksums = {}

    for ti, (time, thead) in enumerate(zip(times, theads)):
        for propk in thead.dtype.names:
            if propk in ('NX', 'NY', 'NZ'):
                thead[propk] = '%3d' % props[propk]
            elif propk == 'LENH':
                thead[propk] = '%4d' % datamap['vardef'][ti].itemsize
            else:
                thead[propk] = getattr(infile, propk)
        timestr = time.strftime('%y%m%d%H').encode('ascii') + FF
        thead['YYMMDDHHFF'] = timestr
        for sfck in sfckeys:
            invar = infile.variables[sfck.decode()]
            var_time = datamap['surface'][sfck.decode()]
            varhead = var_time['head']

            _skipprop = ('YYMMDDHHFF', 'LEVEL', 'EXP', 'PREC', 'VAR1')
            for varpropk in varhead.dtype.names:
                if varpropk not in _skipprop:
                    varhead[varpropk][ti] = getattr(invar, varpropk)

            indata = invar[ti]
            CVAR, PREC, NEXP, VAR1, KSUM = pack2d(indata, verbose=False)

            varhead['YYMMDDHHFF'][ti] = timestr
            varhead['LEVEL'][ti] = '%2d' % 0
            varhead['PREC'][ti] = '%14.7E' % PREC
            varhead['EXP'][ti] = '%4d' % NEXP
            varhead['VAR1'][ti] = '%14.7E' % VAR1
            checksums[vglvls[0], sfck] = KSUM
            var_time['data'][ti] = CVAR
        for layk in laykeys:
            invar = infile.variables[layk.decode()]
            var_time = datamap['layers'][layk.decode()][ti]
            for li, var_time_lay in enumerate(var_time):
                varhead = var_time_lay['head']
                for varpropk in varhead.dtype.names:
                    if varpropk not in _skipprop:
                        varhead[varpropk] = getattr(invar, varpropk)

                indata = invar[ti, li]
                CVAR, PREC, NEXP, VAR1, KSUM = pack2d(indata)

                varhead['YYMMDDHHFF'] = timestr
                varhead['LEVEL'] = '%2d' % (li + 1)
                var_time_lay['data'] = CVAR
                varhead['PREC'] = '%14.7E' % PREC
                varhead['EXP'] = '%4d' % NEXP
                varhead['VAR1'] = '%14.7E' % VAR1
                vglvl = vglvls[li + 1]
                checksums[vglvl, layk] = KSUM

        keys = {vglvls[0]: sfckeys}
        for vglvl in vglvls[1:]:
            keys[vglvl] = laykeys
        vardef = writevardef(vglvls, keys, checksums)

        datamap['vardef'][ti] = ' '.ljust(datamap['vardef'][ti].itemsize)
        datamap['hdr'][ti] = ' '.ljust(datamap['hdr'][ti].itemsize)
        datamap['vardef'][ti] = vardef.encode('ascii')
        thead['LENH'] = datamap['vardef'][ti].itemsize

    datamap.flush()


def maparlpackedbit(path, mode='r', shape=None, props=None):
    """
    pg 4-9 of http://niwc.noaa.inel.gov/EOCFSV/hysplit/hysplituserguide.pdf

    path - path to existing arl packed bit file or location for new file
    props - if props is empty, the arl path must exist. Otherwise:
        required:
            - NZ integer number of layers including surface
            - NY integer number of rows
            - NX integer number of cols
            - sfckeys variable names in the surface layer
            - laykeys variable names in the layers above the surface
        optional:
            - vglvls list nz of vertical coordinates level
            - checksums dictionary {(vglvl, varkey): integer checksum, ...}

    FILE FORMAT:
        recl = 50 + NX*NY
        for time in times:
            1 rec of length recl with meta-data
                (thdtype + vardefdtype + hdrdtype)
            for sfckey in sfckeys:
                1 rec of length recl (vhdtype + nx*ny bytes)
            for layer in layers:
                for laykey in laykeys:
                    1 rec of length recl (vhdtype + nx*ny bytes)
    """
    if props is None:
        props = {}

    if props == {}:
        props.update(inqarlpackedbit(path))
    else:
        srflen = 6 + 2 + (4 + 3 + 1) * len(props['sfckeys'])
        laylen = (6 + 2 + (4 + 3 + 1) *
                  len(props['laykeys'])) * (props['NZ'] - 1)
        props['LENH'] = 108 + srflen + laylen

    nx = props['NX']
    ny = props['NY']
    # minus 1 excludes surface
    # nz = props['NZ'] - 1
    hlen = props['LENH']
    sfckeys = props['sfckeys']
    laykeys = props['laykeys']
    ncell = nx * ny
    vardefdtype = dtype('>S%d' % hlen)
    hdrdtype = dtype('>S%d' % (50 + ncell - hlen - thdtype.itemsize))
    lay1dtype = dtype(
        [('head', vhdtype), ('data', dtype('(%d,%d)>1S' % (ny, nx)))])
    sfcdtype = dtype(dict(names=[k.decode() for k in sfckeys], formats=[
                     lay1dtype] * len(sfckeys)))
    layersdtype = dtype([(str(laykey),
                          dtype(dict(names=[k.decode()
                                            for k in layvarkeys],
                                     formats=[lay1dtype] * len(layvarkeys))))
                         for laykey, layvarkeys in laykeys])
    timedtype = dtype([('timehead', thdtype), ('vardef', vardefdtype),
                       ('hdr', hdrdtype), ('surface', sfcdtype),
                       ('layers', layersdtype)])
    datamap = np.memmap(path, timedtype, shape=shape, mode=mode)
    return datamap


def unpack(bytes, VAR1, EXP):
    """
    bytes - nx by ny bytes
    VAR1 - as read directly from LABEL as BYTES with dimension time
    EXP - EXP as read directly from LABEL as BYTES with dimension time
    """
    vold = VAR1.astype('f')
    scale = np.float32(2.0)**np.float32(7 - EXP.astype('i'))
    invscale = np.float32(1.) / scale
    data = (bytes.view('uint8') - np.float32(127.)) * invscale[..., None, None]
    data[..., 0, 0] += vold
    data[..., 0] = np.cumsum(data[..., 0], axis=data.ndim - 2)
    vdata = np.cumsum(data, axis=data.ndim - 1)
    return vdata


def CHAR(c):
    return chr(c).encode('ascii')


def pack2d(RVARA, verbose=False):
    """
    CHARACTER, INTENT(OUT) :: cvar(nxy) ! packed char*1 output array
    REAL,      INTENT(OUT) :: prec      ! precision of packed data array
    INTEGER,   INTENT(OUT) :: nexp      ! packing scaling exponent
    REAL,      INTENT(OUT) :: var1      ! value of real array at position (1,1)
    INTEGER,   INTENT(OUT) :: ksum      ! rotating checksum of packed data
    """
    # MAX = np.maximum
    FLOAT = np.float32
    INT = np.int32
    # ABS = np.abs
    LOG = np.log
    NY, NX = RVARA.shape
    CVAR = np.zeros(RVARA.shape, dtype='uint8')
    RVAR = RVARA.astype('f')
    VAR1 = RVAR[0, 0]

    # find the maximum difference between adjacent elements
    # START ORIGINAL SERIAL CODE
    # ROLD= VAR1
    # RMAX= 0.0
    # for J in range(NY):
    #    for I in range(NX):
    #        # compute max difference between elements along row
    #        RMAX=MAX( ABS(RVAR[J, I]-ROLD), RMAX)
    #        ROLD=RVAR[J,I]
    #
    #    # row element 1 difference always from previous row
    #    ROLD=RVAR[J, 0]
    #
    # END ORIGINAL SERIAL CODE
    # START NUMPY VECTOR CODE
    colmax = np.abs(np.diff(RVAR, axis=1)).max()
    rowmax = np.abs(np.diff(np.append(VAR1, RVAR[:, 0]), axis=0)).max()
    RMAX = np.maximum(colmax, rowmax)
    # END NUMPY VECTOR CODE

    SEXP = 0.0
    # compute the required scaling exponent
    if RMAX != 0.0:
        SEXP = LOG(RMAX) / LOG(np.float32(2.))

    NEXP = INT(SEXP)
    # positive or whole number scaling round up for lower precision
    if SEXP >= 0.0 or (SEXP % 1.0) == 0.0:
        NEXP = NEXP + 1
    # precision range is -127 to 127 or 254
    PREC = np.float32((2.0**NEXP) / 254.0)
    SCEXP = np.float32(2.0**(7 - NEXP))

    # START ORIGINAL SERIAL CODE
    # # initialize checksum
    # KSUM=0
    #
    # K=0
    # # set column1 value
    # RCOL=VAR1
    # RCOL = VAR1
    #
    # # pack the array from low to high
    # for J in range(NY):
    #     ROLD = RCOL
    #     for I in range(NX):
    #         K=K+1
    #         # packed integer at element
    #         ICVAL=INT((RVAR[J, I]-ROLD)*SCEXP+127.5)
    #
    #         # convert to character
    #         #CVAR[J, I]=CHAR(ICVAL)
    #         CVAR[J, I]=ICVAL
    #         # previous element as it would appear unpacked
    #         ROLD=FLOAT(ICVAL-127)/SCEXP+ROLD
    #         # save the first column element for next row
    #         if I == 0:
    #             RCOL=ROLD.copy()
    #         # maintain rotating checksum
    #         KSUM=KSUM+ICVAL
    #         # if sum carries over the eighth bit add one
    #         if KSUM >= 256:
    #             KSUM=KSUM-255
    #
    # CVART = CVAR.copy()
    # KSUMT = KSUM
    # END ORIGINAL SERIAL CODE

    # START NUMPY VECTOR CODE
    ROLDS = np.zeros_like(RVAR[:, 0])
    ROLD = VAR1
    for J in range(NY):
        ICVAL = INT((RVAR[J, 0] - ROLD) * SCEXP + 127.5)
        CVAR[J, 0] = ICVAL
        ROLD = FLOAT(ICVAL - 127) / SCEXP + ROLD
        ROLDS[J] = ROLD

    ROLD = ROLDS
    for I in range(1, NX):
        ICVAL = INT((RVAR[:, I] - ROLD) * SCEXP + 127.5)
        CVAR[:, I] = ICVAL
        ROLD = FLOAT(ICVAL - 127) / SCEXP + ROLD
    KSUM = INT(CVAR.sum()) % 255
    # END NUMPY VECTOR CODE

    # assert((CVART == CVAR).all())
    # assert(KSUM == KSUMT)
    return CVAR.view('>S1'), PREC, NEXP, VAR1, KSUM


class arlpackedbit(PseudoNetCDFFile):
    """
    arlpackedbit reads files formatted according to NOAA's arl packed bit
    format

    Format as follows:

    for t in times:
        thdtype = dtype([('YYMMDDHHFF', '>10S'), ('LEVEL', '>2S'),
                         ('GRID', '>2S') , ('INDX', '>4S'), ('Z1', '>4S'),
                         ('MB', '2>14S'),  ('SRCE', '>S4'), ('MGRID', '>S3'),
                         ('VSYS', '>S2'), ('POLLAT', '>S7'), ('POLLON', '>S7'),
                         ('REFLAT','>S7'), ('REFLON','>S7'), ('GRIDX','>S7'),
                         ('ORIENT','>S7'), ('TANLAT','>S7'), ('SYNCHX','>S7'),
                         ('SYNCHY','>S7'), ('SYNCHLAT','>S7'),
                         ('SYNCHLON','>S7'), ('RESERVED', '>S7'),
                         ('NX', '>S3'), ('NY', '>S3'), ('NZ', '>S3'),
                         ('VSYS2', '>S2'), ('LENH', '>S4')]) +
                         ny * nx - 158 - 50
        ny+nx-
        for surfacekey in surfacekeys:
            YYMMDDHHFFLLGGVVVVEEEEPPPPPPPPPPPPPPVVVVVVVVVVVVVV+
            P(ny,nx)
        for layerkey in layerkeys:
            for layer in layers:
                YYMMDDHHFFLLGGVVVVEEEEPPPPPPPPPPPPPPVVVVVVVVVVVVVV+
                P(ny,nx)


    YY = %y in strftime for GMT date
    MM = %m in strftime
    DD = %d in strftime
    HH = %H of model hour
    FF = %H of forecast hour
    LL = 2-digit layer
    GG = 2-digit grid
    IIII = 'INDX'
    ZZZZ = 4-digit integer 0
    XXXXXXXXXXXXX = %14.7e 0
    YYYYYYYYYYYYY = %14.7e 0
    EEEE = 4-digit integer exponent
    PPPPPPPPPPPPP = %14.7e precision
    VVVVVVVVVVVVV = %14.7e value R(1,1)
    P(ny,nx) = ny by nx 1-byte values dervied from real values (R) according
               to the formula below.

    P(i,j) = (Ri,j  - Ri-1,j)* (2**(7-(ln dRmax / ln 2)))
    """
    @classmethod
    def isMine(cls, path):
        testchunk = np.fromfile(path,
                                dtype=dtype([('YYMMDDHHFF', '>10S'),
                                             ('LEVEL', '>2S'), ('GRID', '>2S'),
                                             ('INDX', '>4S')]), count=1)
        check = testchunk[0]['INDX'] == b'INDX'
        return check

    def __init__(
        self, path, shape=None, cache=False,
        earth_radius=6370000, synchlon=None, synchlat=None
    ):
        """
        Parameters
        ----------
        path : string
            path to arl packed bit formatted file
        shape : tuple or None
            shape of file to over ride derived shape
        earth_radius : scalar
            radius of the earth used in converting projected to lat/lon
        synchlon: scalar
            The SYNCHLON variable has 6 digits of precision in the file
            this keyword allows you to provide more.
        synchlat: scalar
            The SYNCHLAT variable has 6 digits of precision in the file
            this keyword allows you to provide more.

        Returns
        -------
        arlf : arlpackedbit
            PseudoNetCDF file with packed bit contents unpacked
        """
        self._path = path
        self._f = f = open(path, 'r')
        self._cache = cache
        self._earth_radius = earth_radius
        # f.seek(0, 2)
        # fsize = f.tell()
        f.seek(0, 0)

        # vord = np.vectorize(ord)
#        timesfc = []
#        times = []
#        tflag = []

        f.seek(0, 0)
        props = {}
        self._datamap = maparlpackedbit(path, shape=shape, props=props)
        datamap = self._datamap
        t0hdr = datamap['timehead'][0]
        for key in thdtype.names:
            setattr(self, key, t0hdr[key])

        out = readvardef(datamap['vardef'][0])
        alllayvarkeys = []
        for layk, layvarkeys in out['laykeys']:
            alllayvarkeys.extend(
                [k.decode() for k in layvarkeys
                 if k.decode() not in alllayvarkeys])
        self._layvarkeys = tuple(alllayvarkeys)
        tflag = np.char.replace(datamap['timehead']['YYMMDDHHFF'], b' ', b'0')
        sfckeys = self._datamap['surface'].dtype.names
        laykeys = self._layvarkeys

        self.variables = PseudoNetCDFVariables(
            func=self._getvar, keys=list(sfckeys + laykeys))
        self.createDimension('time', datamap.shape[0])
        # minus 1 excludes surface
        self.createDimension('z', int(self.NZ) - 1)
        if synchlon is None:
            self._synchlon = float(self.SYNCHLON)
        else:
            self._synchlon = synchlon

        if synchlat is None:
            self._synchlat = float(self.SYNCHLAT)
        else:
            self._synchlat = synchlat

        gridx = float(t0hdr['GRIDX']) * 1000.
        nx = props['NX']
        ny = props['NY']
        if gridx != 0:
            self.XSIZE = self.YSIZE = gridx
            self.createDimension('x', nx)
            self.createDimension('y', ny)
            x = self.createVariable('x', 'f', ('x',))
            x[:] = np.arange(x[:].size) * gridx
            y = self.createVariable('y', 'f', ('y',))
            y[:] = np.arange(y[:].size) * gridx
            self._addcrs()
        else:
            self.createDimension('x', nx)
            self.createDimension('y', ny)
            self.createDimension('nv', 2)
            x = self.createVariable('x', 'f', ('x',))
            x.standard_name = 'longitude'
            x.units = 'degrees'
            xe = self.createVariable('x_bounds', 'f', ('x', 'nv'))
            xe.standard_name = 'longitude_bounds'
            xe.units = 'degrees'
            x[:] = np.arange(0, nx) * float(self.REFLON) + self._synchlon
            xe[:-1, 1] = x[:1] + np.diff(x) / 2
            xe[1:, 0] = x[1:] - np.diff(x) / 2
            xe[0, 0] = x[0] - np.diff(x)[0] / 2
            xe[1, 1] = x[1] + np.diff(x)[1] / 2
            y = self.createVariable('y', 'f', ('y',))
            y.standard_name = 'latitude'
            y.units = 'degrees'
            ye = self.createVariable('y_bounds', 'f', ('y', 'nv'))
            ye.standard_name = 'latitude_bounds'
            ye.units = 'degrees'
            y[:] = np.arange(0, ny) * float(self.REFLAT) + self._synchlat
            ye[:-1, 1] = y[:1] + np.diff(y) / 2
            ye[1:, 0] = y[1:] - np.diff(y) / 2
            ye[0, 0] = y[0] - np.diff(y)[0] / 2
            ye[1, 1] = y[1] + np.diff(y)[1] / 2

        times = [datetime.strptime(
            t.astype('S8').decode(), '%y%m%d%H') for t in tflag]
        rtime = times[0]
        hours_since = [(t - rtime).total_seconds() // 3600 for t in times]
        timev = self.createVariable('time', 'i', ('time',))
        timev[:] = hours_since
        timev.units = rtime.strftime('hours since %F %H:%M:%S')
        z = self.createVariable('z', 'f', ('z',))
        z[:] = out['vglvls'][1:]
        self.SFCVGLVL = out['vglvls'][0]
        _units = {1: 'pressure sigma', 2: 'pressure absolute',
                  3: 'terrain sigma', 4: 'hybrid sigma'}
        z.units = _units.get(int(self.VSYS2), 'unknown')
        self.setCoords(['time', 'z', 'y', 'x'])

    def _addcrs(self):
        x = self.variables['x']
        y = self.variables['y']
        gridx = np.diff(x[:])[0]
        gridy = np.diff(y[:])[0]
        nx = x.size
        ny = y.size
        crs = self.createVariable('crs', 'i', ())
        s = self
        tanlat = np.float64(s.TANLAT)
        reflon = np.float64(s.REFLON)
        reflat = np.float64(s.REFLAT)
        atanlat = np.abs(tanlat)
        pname = crs.grid_mapping_name = {
            0: 'equatorial_mercator',
            90: 'polar_stereographic',
        }.get(atanlat, 'lambert_conformal_conic')
        if pname == 'lambert_conformal_conic':
            crs.standard_parallel = np.array([tanlat, tanlat])
            crs.longitude_of_central_meridian = reflon
            crs.latitude_of_projection_origin = reflat
        elif pname == 'polar_stereographic':
            crs.straight_vertical_longitude_from_pole = reflon
            crs.standard_parallel = reflat
            crs.latitude_of_projection_origin = tanlat
        else:
            raise KeyError('Not yet implemented equatorial mercator')

        crs.earth_radius = self._earth_radius  # WRF-based radius
        scellx = (float(self.SYNCHX) - 1) * gridx  # 1-based I cell
        scelly = (float(self.SYNCHY) - 1) * gridy  # 1-based J cell
        slon, slat = self._synchlon, self._synchlat
        if slon > 180:
            slon = slon % 180 - 180

        halfwidth = gridx * (nx - 1) / 2
        halfheight = gridy * (ny - 1) / 2
        crs.false_easting = halfwidth
        crs.false_northing = halfheight
        llcrnrlon, llcrnrlat = self.xy2ll(scellx, scelly)

        x_within_precision = np.round(slon / llcrnrlon, 3) == 1
        y_within_precision = np.round(slat / llcrnrlat, 4) == 1
        if not (x_within_precision and y_within_precision):
            warn(
                'Grid not centered; using SYNCHLAT/SYNCHLON ' +
                'with limited to 6 significant digits ' +
                'to calculate false easting/northing'
            )
            crs.false_easting = 0.
            crs.false_northing = 0.
            llcrnrx, llcrnry = self.ll2xy(slon, slat)
            crs.false_easting = -llcrnrx + scellx
            crs.false_northing = -llcrnry + scelly

        self.Conventions = 'CF-1.6'

    def _getvar(self, varkey):
        datamap = self._datamap
        stdname, stdunit = stdprops.get(varkey, (varkey, 'unknown'))
        if varkey in datamap['surface'].dtype.names:
            vhead = datamap['surface'][varkey]['head']
            v11 = vhead['VAR1']
            EXP = vhead['EXP']
            props = dict([(pk, vhead[pk][0]) for pk in vhead.dtype.names
                          if pk not in ('YYMMDDHHFF', 'LEVEL')])
            bytes = datamap['surface'][varkey]['data']
            vdata = unpack(bytes, v11, EXP)
            # if varkey == 'MXLR':
            #  CVAR, PREC, NEXP, VAR1, KSUM = pack2d(vdata[21], verbose = True)

            out = PseudoNetCDFVariable(
                self, varkey, 'f', ('time', 'y', 'x'),
                values=vdata, units=stdunit, standard_name=stdname, **props
            )
        elif varkey in self._layvarkeys:
            laykeys = datamap['layers'].dtype.names
            mylaykeys = [laykey for laykey in laykeys
                         if varkey in datamap['layers'][laykey].dtype.names]
            bytes = np.array([datamap['layers'][lk][varkey]['data']
                              for lk in mylaykeys]).swapaxes(0, 1)
            vhead = np.array([datamap['layers'][lk][varkey]['head']
                              for lk in mylaykeys]).swapaxes(0, 1)
            v11 = vhead['VAR1']
            EXP = vhead['EXP']
            props = dict([(pk, vhead[pk][0, 0]) for pk in vhead.dtype.names
                          if pk not in ('YYMMDDHHFF', 'LEVEL')])
            props['LEVEL_START'] = vhead['LEVEL'][0, 0]
            props['LEVEL_END'] = vhead['LEVEL'][-1, -1]
            vdata = unpack(bytes, v11, EXP)
            out = PseudoNetCDFVariable(
                self, varkey, 'f', ('time', 'z', 'y', 'x'),
                values=vdata, units=stdunit,
                standard_name=stdname, **props
            )
        if self._cache:
            self.variables[varkey] = out
        return out

    def getMap(self, maptype='basemap_auto', **kwds):
        if 'latitude_bounds' in self.variables:
            return PseudoNetCDFFile.getMap(self, maptype=maptype, **kwds)

        from PseudoNetCDF.coordutil import basemap_from_proj4
        kwds = kwds.copy()
        myproj = self.getproj(withgrid=True, projformat='pyproj')
        myprojstr = self.getproj(withgrid=True, projformat='proj4')
        llcrnrlon, llcrnrlat = myproj(
            self.variables['x'][0], self.variables['y'][0], inverse=True
        )
        urcrnrlon, urcrnrlat = myproj(
            self.variables['x'][-1], self.variables['y'][-1], inverse=True
        )
        kwds['llcrnrlon'] = llcrnrlon
        kwds['llcrnrlat'] = llcrnrlat
        kwds['urcrnrlon'] = urcrnrlon
        kwds['urcrnrlat'] = urcrnrlat
        return basemap_from_proj4(myprojstr, **kwds)

    def plot(self, *args, **kwds):
        """
        See PseudoNetCDF.PseudoNetCDFFile.plot. For this file,
        the default plottype is 'x-y'
        """
        kwds.setdefault('plottype', 'x-y')
        ax = PseudoNetCDFFile.plot(self, *args, **kwds)
        if kwds['plottype'] == 'x-y':
            map_kw = kwds.get('map_kw', None)
            if map_kw is False:
                return ax
            elif map_kw is None or map_kw is True:
                map_kw = {}

            try:
                map_kw = map_kw.copy()
                coastlines = map_kw.pop('coastlines', True)
                countries = map_kw.pop('countries', True)
                states = map_kw.pop('states', False)
                counties = map_kw.pop('counties', False)
                bmap = self.getMap(**map_kw)
                if coastlines:
                    bmap.drawcoastlines(ax=ax)
                if countries:
                    bmap.drawcountries(ax=ax)
                if states:
                    bmap.drawstates(ax=ax)
                if counties:
                    bmap.drawcounties(ax=ax)
            except Exception:
                pass
        return ax

    def getproj(self, withgrid=False, projformat='pyproj'):
        from PseudoNetCDF.coordutil import getproj4_from_cf_var
        gridmapping = self.variables['crs']
        proj4str = getproj4_from_cf_var(gridmapping, withgrid=withgrid)
        preserve_units = withgrid
        if projformat == 'proj4':
            return proj4str
        elif projformat == 'pyproj':
            import pyproj
            # pyproj adds +units=m, which is not right for latlon/lonlat
            if '+proj=lonlat' in proj4str or '+proj=latlon' in proj4str:
                preserve_units = True
            return pyproj.Proj(proj4str, preserve_units=preserve_units)
        elif projformat == 'wkt':
            import osr
            srs = osr.SpatialReference()
            # Imports WKT to Spatial Reference Object
            srs.ImportFromProj4(proj4str)
            srs.ExportToWkt()  # converts the WKT to an ESRI-compatible format
            return srs.ExportToWkt()
        else:
            raise ValueError('projformat must be pyproj, proj4 or wkt')


registerwriter('noaafiles.arlpackedbit', writearlpackedbit)
registerwriter('arlpackedbit', writearlpackedbit)

if __name__ == '__main__':
    import sys
    out = arlpackedbit(sys.argv[1])



from __future__ import print_function
import sys
from warnings import warn
from types import MethodType
from PseudoNetCDF.netcdf import NetCDFVariable
from .sci_var import PseudoNetCDFFile
from .sci_var import get_ncf_object
import numpy as np
from PseudoNetCDF.camxfiles import Writers as CAMxWriters
import PseudoNetCDF.geoschemfiles as geoschemwriters
import PseudoNetCDF.icarttfiles.ffi1001 as icarttwriters


if sys.version_info > (3,):
    long = int


class Pseudo2NetCDF:
    """
    Pseudo2NetCDF is a base class for conversion.  Properties and methods can
    be overwritten to facilitate conversion of special PseudoNetCDFFiles.

    Specifically: ignore_global_properties and ignore_variable_properties lists
    can be overwritten so that class properties and methods are not written
    to a netCDF file
    """
    import re
    ignore_global_re = re.compile('^_\w*(__\w*)?')
    ignore_variable_re = re.compile('^_\w*(__\w*)?')
    ignore_global_properties = ['variables', 'dimensions']
    ignore_variable_properties = ['typecode', 'dimensions']
    special_properties = ['_fillvalue', '_FillValue']
    unlimited_dimensions = []
    create_variable_kwds = {}

    def __init__(self, datafirst=False, verbose=1):
        self.datafirst = datafirst
        self.verbose = verbose

    def convert(self, pfile, npath=None, inmode='r', outmode='w',
                format='NETCDF4'):
        pfile = get_ncf_object(pfile, inmode)
        nfile = get_ncf_object(npath, outmode, format=format)
        if self.verbose:
            print("Adding dimensions", file=sys.stdout)
        self.addDimensions(pfile, nfile)
        if self.verbose:
            print("Adding globals", file=sys.stdout)
        self.addGlobalProperties(pfile, nfile)
        if self.verbose:
            print("Adding variables", file=sys.stdout)
        self.addVariables(pfile, nfile)
        nfile.sync()
        return nfile

    def addDimensions(self, pfile, nfile):
        for d in pfile.dimensions.keys():
            self.addDimension(pfile, nfile, d)

    def addDimension(self, pfile, nfile, d):
        v = pfile.dimensions[d]
        unlim = (d in self.unlimited_dimensions or v.isunlimited())
        if not isinstance(v, (int, long)) and v is not None:
            v = len(v)

        if unlim:
            if isinstance(nfile, PseudoNetCDFFile):
                nd = nfile.createDimension(d, v)
                nd.setunlimited(True)
            else:
                nd = nfile.createDimension(d, None)
        else:
            nd = nfile.createDimension(d, v)

        nfile.sync()

    def addGlobalProperties(self, pfile, nfile):
        for k in [k for k in pfile.ncattrs()
                  if (k not in self.ignore_global_properties and
                      self.ignore_global_re.match(k) is None)]:
            value = getattr(pfile, k)
            if not isinstance(value, MethodType):
                try:
                    setattr(nfile, k, value)
                except TypeError as e:
                    if isinstance(value, bool):
                        setattr(nfile, k, np.int8(value))
                    else:
                        raise e
                except Exception as e:
                    warn("Could not add %s to file; %s: %s" % (k, type(e), e))

    def addVariableProperties(self, pvar, nvar):
        for a in [k for k in pvar.ncattrs()
                  if ((k not in self.ignore_variable_properties and
                       self.ignore_variable_re.match(k) is None) or
                      k in self.special_properties)]:
            value = getattr(pvar, a)
            if isinstance(nvar, NetCDFVariable) and a == '_FillValue':
                continue
            if not isinstance(value, MethodType):
                try:
                    nvar.setncattr(a, value)
                    # setattr(nvar,a,value)
                except TypeError as e:
                    if isinstance(value, bool):
                        nvar.setncattr(a, np.int8(value))
                    else:
                        raise e
                except Exception as e:
                    if 'long_name' in pvar.ncattrs():
                        warn("Could not add %s=%s to variable %s; %s" %
                             (a, str(value), str(pvar.long_name), e))
                    else:
                        warn("Could not add %s=%s to variable; %s" %
                             (a, str(value), e))

    def addVariable(self, pfile, nfile, k, data=True):
        pvar = pfile.variables[k]
        try:
            typecode = pvar.typecode()
        except Exception:
            typecode = pvar[...].dtype.char

        create_variable_kwds = self.create_variable_kwds.copy()
        if hasattr(pvar, 'missing_value'):
            create_variable_kwds['fill_value'] = pvar.missing_value
        elif hasattr(pvar, 'fill_value'):
            create_variable_kwds['fill_value'] = pvar.fill_value
        elif hasattr(pvar, '_FillValue'):
            create_variable_kwds['fill_value'] = pvar._FillValue

        nvar = nfile.createVariable(
            k, typecode, pvar.dimensions, **create_variable_kwds)
        self.addVariableProperties(pvar, nvar)
        if data:
            self.addVariableData(pfile, nfile, k)
            nfile.sync()

        try:
            nfile.flush()
        except Exception:
            pass
        del pvar, nvar

    def addVariableData(self, pfile, nfile, k):
        from numpy.ma import MaskedArray
        from numpy import isscalar
        nvar = nfile.variables[k]
        pvar = pfile.variables[k]
        if isscalar(nvar) or nvar.ndim == 0:
            if isinstance(pvar, NetCDFVariable):
                pvar = pvar[...]
            nvar[...] = pvar
        elif isinstance(pvar[...], MaskedArray):
            nvar[:] = pvar[...].filled(getattr(nvar, 'fill_value', getattr(
                nvar, '_FillValue', getattr(pvar, 'missing_value', -9999))))
        else:
            nvar[:] = pvar[...]

    def addVariables(self, pfile, nfile):
        for k in pfile.variables.keys():
            if self.verbose:
                print("Defining", k, file=sys.stdout)
            self.addVariable(pfile, nfile, k, data=self.datafirst)

        if self.datafirst:
            nfile.sync()

        if not self.datafirst:
            for k in pfile.variables.keys():
                if self.verbose:
                    print("Populating", k, file=sys.stdout)
                self.addVariableData(pfile, nfile, k)
            nfile.sync()


def pywriter(ifile, outpath, data=True):
    print("""# Import Libraries and Functions
from netCDF4 import Dataset
from numpy import *
from numpy.ma import masked_array

# Open file for writing
outpath = '%s'
outfile = Dataset(outpath, 'w')

""" % outpath)
    print("# Define Dimensions")
    for dk, dv in ifile.dimensions.items():
        dl = len(dv)
        if dv.isunlimited():
            dl = None
        print("dim_%s = outfile.createDimension('%s', %s); # %d" %
              (dk, dk, dl, len(dv)))

    print("# Add global properties")
    for pk in ifile.ncattrs():
        pv = getattr(ifile, pk)
        if isinstance(pv, bool):
            pv = int(pv)
        print("setattr(outfile, '%s', %s)" % (pk, repr(pv)))

    print("## Define Variables")
    print("vars = {}")
    for vk, v in ifile.variables.items():
        print("# Defining " + vk)
        print("var = vars['%s'] = outfile.createVariable('%s', '%s', %s)" % (
            vk, vk, v.dtype.char, v.dimensions))
        for pk in v.ncattrs():
            pv = getattr(v, pk)
            print("setattr(var, '%s', %s)" % (pk, repr(pv)))
        print("")

    if data:
        print("## Populate Variables")
        for vk, v in ifile.variables.items():
            print("# Populating " + vk)
            if isinstance(v, np.ma.MaskedArray):
                vtype = np.ma.MaskedArray
            else:
                vtype = np.ndarray
            print("var = vars['%s']" % vk)
            print("var[:] = %s" % (repr(v[:].view(type=vtype))))


def pncgen(ifile, outpath, inmode='r', outmode='w', format='NETCDF4_CLASSIC',
           verbose=1, complevel=0, writer_kw=None):
    """
    ifile - input file to write out
    outpath - path to outputfile
    inmode - how is file read (if ifile is a path)
    outmode - w, w+s
    format - any PseudoNetCDF or Dataset option
    """
    if writer_kw is None:
        writer_kw = {}
    if format[:6] == 'NETCDF':
        p2n = Pseudo2NetCDF()
        p2n.verbose = verbose
        if complevel > 0:
            p2n.create_variable_kwds['zlib'] = True
            p2n.create_variable_kwds['complevel'] = complevel
        return p2n.convert(ifile, outpath, inmode=inmode, outmode=outmode,
                           format=format)

    from ._getwriter import getwriterdict
    writerdict = getwriterdict()
    if format == 'python':
        pywriter(ifile, outpath)
    elif format == 'csv':
        from .textfiles._delimited import ncf2csv
        ncf2csv(ifile, outpath)
    elif format in writerdict:
        writer = writerdict[format]
        return writer(ifile, outpath, **writer_kw)
    else:
        for writers in [CAMxWriters, geoschemwriters, icarttwriters]:
            writer = getattr(writers, 'ncf2%s' % format, None)
            if writer is not None:
                return writer(ifile, outpath, **writer_kw)
                break
        else:
            raise KeyError('Unknown output file type "%s"' % format)


def main():
    from .pncparse import pncparse
    ifiles, options = pncparse(has_ofile=True, interactive=False)
    if len(ifiles) != 1:
        raise IOError(
            'pncgen can output only 1 file; user requested %d' % len(ifiles))
    ifile, = ifiles
    return pncgen(ifile, options.outpath, outmode=options.mode,
                  format=options.outformat, verbose=options.verbose), options


if __name__ == '__main__':
    main()



__doc__ = """
.. _uamiv
:mod:`uamiv` -- UAM-IV File Interfaces
======================================

.. module:: uamiv
   :platform: Unix, Windows
   :synopsis: Provides :ref:`PseudoNetCDF` memory map and random access read
   based file interfaces for CAMx UAM-IV files.
.. moduleauthor:: Barron Henderson <barronh@unc.edu>
"""
__all__ = ['Memmap', 'Read', 'Write', 'Transforms']

from . import Memmap
from . import Read
from . import Write
from . import Transforms

# _camx_units is based on file name and an aerosol flag (True = aerosol)
if __name__ == '__main__':
    from PseudoNetCDF.camxfiles.uamiv.Memmap import uamiv
    from PseudoNetCDF.pncdump import pncdump_parser, \
        dump_from_cmd_line
    parser = pncdump_parser()
    (file_path, options, extra_args_dict) = parser.parse_args()

    dump_from_cmd_line(file_path, options, uamiv)



from __future__ import unicode_literals
from tempfile import NamedTemporaryFile as tnf

from ._files import PseudoNetCDFFile


def get_ncf_object(path_or_object, mode, format='NETCDF4_CLASSIC'):
    """
    Return an open NetCDF or PseudoNetCDF from an object or path
    """
    from os.path import exists, isfile, isdir
    from PseudoNetCDF.netcdf import NetCDFFile
    read_only = ('r', 'r+', 'rs', 'rs+', 'r+s')
    if isinstance(path_or_object, (str, )):
        if exists(path_or_object):
            if isfile(path_or_object):
                ncf_object = NetCDFFile(path_or_object, mode, format=format)
            elif isdir(path_or_object):
                raise ValueError(("Got directory at %s; not sure what " +
                                  "to do") % path_or_object)
            else:
                raise ValueError(
                    "Expected file or directory at %s" % path_or_object)
        elif mode not in read_only:
            ncf_object = NetCDFFile(path_or_object, mode, format=format)
        else:
            raise IOError("Cannot open missing file for reading")
    elif (
        isinstance(path_or_object, NetCDFFile) or
        isinstance(path_or_object, PseudoNetCDFFile)
    ):
        return path_or_object
    elif path_or_object is None and mode not in read_only:
        tfile = tnf(mode='w+b')
        npath = tfile.name
        ncf_object = NetCDFFile(npath, mode)
    else:
        raise ValueError("Not a path; not a netCDF file; not a " +
                         "PseudoNetCDF file... I don't know what to do")
    return ncf_object


def get_dimension_length(pfile, key):
    """
    Return the length of a dimension (key) from pfile
    """
    dim = pfile.dimsensions[key]
    if dim is None:
        for k in pfile.variables.keys():
            v = pfile.variables[k]
            if key in v.dimensions:
                return v[...].shape[list(v.dimensions).index(key)]
        return 0
    elif isinstance(dim, int):
        return dim
    else:
        return len(dim)



from PseudoNetCDF.pncwarn import warn
import numpy as np
_withlatlon = False
try:
    import pyproj
    _withlatlon = True
except Exception:
    warn('pyproj could not be found, so IO/API coordinates cannot be ' +
         'converted to lat/lon; to fix, install pyproj or basemap ' +
         '(e.g., `pip install pyproj)`')


def add_lay_coordinates(ifileo):
    if 'LAY' in ifileo.dimensions:
        nlay = len(ifileo.dimensions['LAY'])
    elif hasattr(ifileo, 'NLAYS'):
        nlay = ifileo.NLAYS
    elif hasattr(ifileo, 'VGLVLS'):
        nlay = len(ifileo.VGLVLS) - 1
    else:
        return
    if 'layer' not in ifileo.variables.keys():
        var = ifileo.createVariable('layer', 'd', ('LAY',))
        var[:] = np.arange(nlay, dtype='d')
        var.units = 'model layers'
        var.standard_name = 'layer'
    if 'level' not in ifileo.variables.keys():
        var = ifileo.createVariable('level', 'd', ('LAY',))
        if hasattr(ifileo, 'VGLVLS'):
            var[:] = (ifileo.VGLVLS[:-1] + ifileo.VGLVLS[1:]) / 2
        else:
            var[:] = np.arange(nlay, dtype='d')
        var.units = 'sigma'
        var.positive = 'down'
        var.standard_name = 'level'
        var._CoordinateAxisType = "GeoZ"
        var._CoordinateZisPositive = "down"


def add_time_variables(ifileo):
    add_time_variable(ifileo, 'time')
    add_time_variable(ifileo, 'time_bounds')


def add_time_variable(ifileo, key):
    from datetime import datetime, timedelta, timezone
    strptime = datetime.strptime

    sdate = int(abs(ifileo.SDATE))
    if sdate < 1400000:
        sdate += 2000000
    sdate = datetime(sdate // 1000, 1, 1, tzinfo=timezone.utc) + \
        timedelta(days=(sdate % 1000) - 1)
    sdate = sdate + timedelta(days=ifileo.STIME / 240000.)
    rdate = strptime('1970-01-01 00:00:00+0000', '%Y-%m-%d %H:%M:%S%z')
    if ifileo.TSTEP == 0:
        tmpseconds = 0
    else:
        tmp = ('%06d' % ifileo.TSTEP)
        htmp = tmp[:2]
        mtmp = tmp[2:4]
        stmp = tmp[4:]
        tmpseconds = 3600 * int(htmp) + 60 * int(mtmp) + int(stmp)

    time_unit = "seconds since %s" % (rdate.strftime('%Y-%m-%d %H:%M:%S%z'),)
    if 'TFLAG' in ifileo.variables:
        tflag = ifileo.variables['TFLAG'][:, 0]
        jdays = tflag[:, 0]
        hhmmsss = tflag[:, 1]
        if jdays[0] == 0:
            jdays = [ifileo.SDATE]
        dates = np.array([strptime('%7d %06d+0000' % (jday, hhmmss),
                                   '%Y%j %H%M%S%z')
                          for jday, hhmmss in zip(jdays, hhmmsss)])
        time = np.array([(date - rdate).total_seconds() for date in dates])
        if key == 'time_bounds':
            time = np.array([time, time + tmpseconds]).T
            dims = ('TSTEP', 'tnv')
            if 'tnv' not in ifileo.dimensions.keys():
                ifileo.createDimension('tnv', 2)
        else:
            dims = ('TSTEP',)
    else:
        sdate = strptime('%07d %06d+0000' %
                         (getattr(ifileo, 'SDATE', 1970001),
                          getattr(ifileo, 'STIME', 0)),
                         '%Y%j %H%M%S%z')
        off = (sdate - rdate).total_seconds()
        if key == 'time':
            tmax = max(1, len(ifileo.dimensions['TSTEP']))
            time = np.arange(0, tmax, dtype='i') * tmpseconds + off
            dims = ('TSTEP',)
        elif key == 'time_bounds':
            tmax = max(1, len(ifileo.dimensions['TSTEP'])) + 1
            time = np.arange(0, tmax, dtype='i') * tmpseconds + off
            time = time.repeat(2, 0)[1:-1].reshape(-1, 2)
            dims = ('TSTEP', 'tnv')
            if 'tnv' not in ifileo.dimensions.keys():
                ifileo.createDimension('tnv', 2)
        else:
            raise KeyError(
                'time variables are time and time_bounds, got %s' % key)
    if key not in ifileo.variables.keys():
        var = ifileo.createVariable(key, time.dtype.char, dims)
    else:
        var = ifileo.variables[key]
    var[:] = time

    var.units = time_unit
    if key == 'time':
        var._CoordinateAxisType = "Time"
        var.bounds = 'time_bounds'

    var.long_name = ("synthesized time coordinate from SDATE, " +
                     "STIME, STEP global attributes")


# 0 Clarke 1866
# 1 Clarke 1880
# 2 Bessel
# 3 New International 1967
# 4 International 1909
# 5 WGS 72
# 6 Everest
# 7 WGS 66
# 8 GRS 1980
# 9 Airy
# 10 Modified Everest
# 11 Modified Airy
# 12 WGS 84
# 13 Southeast Asia
# 14 Australian National
# 15 Krassovsky
# 16 Hough
# 17 Mercury 1960
# 18 Modified Mercury 1968
# 19 Normal Sphere
# 20 MM5 Sphere
# 21 WRF-NMM Sphere
_AXIS = np.array([6378206.4, 6378249.145, 6377397.155, 6378157.5, 6378388.0,
                  6378135.0, 6377276.3452, 6378145.0, 6378137.0, 6377563.396,
                  6377304.063, 6377340.189, 6378137.0, 6378155., 6378160.0,
                  6378245.0, 6378270.0, 6378166.0, 6378150.0, 6370997.0,
                  6370000.0, 6371200.0])
_BXIS = np.array([6356583.8, 6356514.86955, 6356078.96284, 6356772.2,
                  6356911.94613, 6356750.519915, 6356075.4133, 6356759.769356,
                  6356752.314140, 6356256.91, 6356103.039, 6356034.448,
                  6356752.314245, 6356773.3205, 6356774.719, 6356863.0188,
                  6356794.343479, 6356784.283666, 6356768.337303, 6370997.0,
                  6370000.0, 6371200.0])


def get_ioapi_sphere():
    import os
    ENV_IOAPI_ISPH = os.environ.get('IOAPI_ISPH', None)
    if ENV_IOAPI_ISPH is None:
        ENV_IOAPI_ISPH = '6370000.'
        warn('IOAPI_ISPH is assumed to be ' +
             ENV_IOAPI_ISPH + '; consistent with WRF')
    isph_parts = [eval(ip) for ip in ENV_IOAPI_ISPH.split(' ')]
    if len(isph_parts) > 2:
        raise ValueError('IOAPI_ISPH must be 1 or 2 parameters (got: %s)' %
                         str(isph_parts))
    elif len(isph_parts) == 2:
        return isph_parts
    elif isph_parts[0] >= 0 and isph_parts[0] < _AXIS.size:
        return _AXIS[isph_parts[0]], _BXIS[isph_parts[0]]
    else:
        return isph_parts * 2


_gdnames = {1: "latitude_longitude", 2: "lambert_conformal_conic",
            7: "mercator", 6: "polar_stereographic"}


def getmapdef(ifileo, add=True):
    gridname = _gdnames[ifileo.GDTYP]
    if add:
        mapdef = ifileo.createVariable(gridname, 'i', ())
    else:
        from PseudoNetCDF import PseudoNetCDFVariable
        mapdef = PseudoNetCDFVariable(ifileo, gridname, 'i', ())
    mapdef.grid_mapping_name = gridname
    if mapdef.grid_mapping_name == "latitude_longitude":
        mapdef.latitude_of_projection_origin = ifileo.YORIG
        mapdef.longitude_of_central_meridian = ifileo.XORIG
    elif mapdef.grid_mapping_name == "polar_stereographic":
        mapdef.latitude_of_projection_origin = ifileo.P_ALP * 90
        mapdef.straight_vertical_longitude_from_pole = ifileo.P_GAM
        mapdef.standard_parallel = np.array([ifileo.P_BET], dtype='d')
        # mapdef.standard_parallel = np.array([ifileo.P_BET], dtype = 'f')
    else:
        mapdef.standard_parallel = np.array(
            [ifileo.P_ALP, ifileo.P_BET], dtype='d')
    if mapdef.grid_mapping_name != "latitude_longitude":
        mapdef.latitude_of_projection_origin = ifileo.YCENT
        mapdef.longitude_of_central_meridian = ifileo.XCENT
        mapdef.false_northing = -ifileo.YORIG
        mapdef.false_easting = -ifileo.XORIG
        mapdef._CoordinateTransformType = "Projection"
        mapdef._CoordinateAxes = "x y"
    ioapi_sphere = get_ioapi_sphere()
    mapdef.semi_major_axis = getattr(ifileo, 'semi_major_axis', getattr(
        ifileo, 'earth_radius', ioapi_sphere[0]))
    mapdef.semi_minor_axis = getattr(ifileo, 'semi_minor_axis', getattr(
        ifileo, 'earth_radius', ioapi_sphere[1]))
    return mapdef


def add_lcc_coordinates(ifileo):
    mapdef = getmapdef(ifileo)
    gridname = mapdef.grid_mapping_name
    if 'PERIM' in ifileo.dimensions.keys():
        xdim = 'PERIM'
        ydim = 'PERIM'
        _x = np.arange(-ifileo.XCELL, (ifileo.NCOLS + 1) * ifileo.XCELL,
                       ifileo.XCELL) + ifileo.XORIG + ifileo.XCELL / 2.
        _y = np.arange(-ifileo.YCELL, (ifileo.NROWS + 1) * ifileo.YCELL,
                       ifileo.YCELL) + ifileo.YORIG + ifileo.YCELL / 2.
        bx = _x[1:]
        by = _y[0].repeat(ifileo.NCOLS + 1)
        ex = _x[-1].repeat(ifileo.NROWS + 1)
        ey = _y[1:]
        tx = _x[0:-1]
        ty = _y[-1].repeat(ifileo.NCOLS + 1)
        wx = _x[0].repeat(ifileo.NROWS + 1)
        wy = _y[:-1]
        lcc_x = x = np.concatenate([bx, ex, tx, wx])
        lcc_y = y = np.concatenate([by, ey, ty, wy])
        XCELL = ifileo.XCELL
        YCELL = ifileo.YCELL
        lcc_xe = np.array([x - XCELL / 2., x + XCELL /
                           2., x + XCELL / 2., x - XCELL / 2.]).T
        lcc_ye = np.array([y - YCELL / 2., y - YCELL /
                           2., y + YCELL / 2., y + YCELL / 2.]).T
        latlon_dim = ('PERIM',)
        latlone_dim = ('PERIM', 'nv')
        latlon_coord = 'PERIM'
    else:
        xdim = 'COL'
        ydim = 'ROW'
        latlon_dim = (ydim, xdim)
        latlon_coord = 'latitude longitude'
        latlone_dim = (ydim, xdim, 'nv')
        xe = np.arange(0, ifileo.NCOLS) * ifileo.XCELL + ifileo.XORIG
        ye = np.arange(0, ifileo.NROWS) * ifileo.YCELL + ifileo.YORIG
        lcc_xe, lcc_ye = np.meshgrid(xe, ye)
        lcc_xe = np.concatenate([lcc_xe[:, :, None],
                                 lcc_xe[:, :, None] + ifileo.XCELL,
                                 lcc_xe[:, :, None] + ifileo.XCELL,
                                 lcc_xe[:, :, None]], axis=2)
        lcc_ye = np.concatenate([lcc_ye[:, :, None], lcc_ye[:, :, None],
                                 lcc_ye[:, :, None] + ifileo.YCELL,
                                 lcc_ye[:, :, None] + ifileo.YCELL], axis=2)
        x = np.arange(0, ifileo.NCOLS) * ifileo.XCELL + \
            ifileo.XCELL / 2. + ifileo.XORIG
        y = np.arange(0, ifileo.NROWS) * ifileo.YCELL + \
            ifileo.YCELL / 2. + ifileo.YORIG
        lcc_x, lcc_y = np.meshgrid(x, y)

    if _withlatlon:
        if ifileo.GDTYP == 2:
            mapstrs = ['+proj=lcc',
                       '+a=%s' % mapdef.semi_major_axis,
                       '+b=%s' % mapdef.semi_minor_axis,
                       '+lon_0=%s' % mapdef.longitude_of_central_meridian,
                       '+lat_1=%s' % mapdef.standard_parallel[0],
                       '+lat_2=%s' % mapdef.standard_parallel[1],
                       '+lat_0=%s' % mapdef.latitude_of_projection_origin]
            mapstr = ' '.join(mapstrs)
            mapproj = pyproj.Proj(mapstr)
        elif ifileo.GDTYP == 6:
            mapstr = ('+proj=stere +a={3} +b={4} ' +
                      '+lon_0={0} +lat_0={1} +lat_ts={2}').format(
                          mapdef.straight_vertical_longitude_from_pole,
                          mapdef.latitude_of_projection_origin,
                          mapdef.standard_parallel[0],
                          mapdef.semi_major_axis,
                          mapdef.semi_minor_axis)
            mapproj = pyproj.Proj(mapstr)
        elif ifileo.GDTYP == 7:
            mapstr = '+proj=merc +a=%s +b=%s +lat_ts=0 +lon_0=%s' % (
                mapdef.semi_major_axis, mapdef.semi_minor_axis,
                mapdef.longitude_of_central_meridian)
            mapproj = pyproj.Proj(mapstr)
        elif ifileo.GDTYP == 1:
            def mapproj(x, y, inverse):
                return (x, y)
        lon, lat = mapproj(lcc_x, lcc_y, inverse=True)
        lone, late = mapproj(lcc_xe.ravel(), lcc_ye.ravel(), inverse=True)
        lone = lone.reshape(*lcc_xe.shape)
        late = late.reshape(*lcc_ye.shape)

    if 'x' not in ifileo.variables.keys() and xdim in ifileo.dimensions:
        """
        Not necessary for cdo
        """
        var = ifileo.createVariable('x', x.dtype.char, (xdim,))
        var[:] = x[:]
        var.units = 'km'
        var._CoordinateAxisType = "GeoX"
        var.long_name = ("synthesized coordinate from XORIG XCELL " +
                         "global attributes")

    if 'y' not in ifileo.variables.keys() and ydim in ifileo.dimensions:
        """
        Not necessary for cdo
        """
        var = ifileo.createVariable('y', x.dtype.char, (ydim,))
        var[:] = y[:]
        var.units = 'km'
        var._CoordinateAxisType = "GeoY"
        var.long_name = ("synthesized coordinate from YORIG YCELL " +
                         "global attributes")

    if _withlatlon and 'latitude' not in ifileo.variables.keys():
        var = ifileo.createVariable('latitude', lat.dtype.char, latlon_dim)
        var[:] = lat
        var.units = 'degrees_north'
        var.standard_name = 'latitude'
        var.bounds = 'latitude_bounds'
        var.coordinates = latlon_coord

    if _withlatlon and 'longitude' not in ifileo.variables.keys():
        var = ifileo.createVariable('longitude', lon.dtype.char, latlon_dim)
        var[:] = lon
        var.units = 'degrees_east'
        var.standard_name = 'longitude'
        var.bounds = 'longitude_bounds'
        var.coordinates = latlon_coord

    if _withlatlon:
        for dk, dl in zip(latlone_dim, late.shape):
            if dk not in ifileo.dimensions:
                ifileo.createDimension(dk, dl)

    if _withlatlon and 'latitude_bounds' not in ifileo.variables.keys():
        var = ifileo.createVariable(
            'latitude_bounds', lat.dtype.char, latlone_dim)
        var[:] = late
        var.units = 'degrees_north'
        var.standard_name = 'latitude_bounds'

    if _withlatlon and 'longitude_bounds' not in ifileo.variables.keys():
        var = ifileo.createVariable(
            'longitude_bounds', lon.dtype.char, latlone_dim)
        var[:] = lone
        var.units = 'degrees_east'
        var.standard_name = 'longitude_bounds'

    for varkey in ifileo.variables.keys():
        var = ifileo.variables[varkey]
        # this must have been a fix for dictionaries that
        # reproduced variables on demand
        # we should find a better fix for this
        # try:
        #    ifileo.variables[varkey] = var
        # except Exception:
        #    pass
        olddims = list(var.dimensions)
        if _withlatlon:
            def io2cf(x):
                return {'ROW': 'latitude', 'COL': 'longitude',
                        'TSTEP': 'time', 'LAY': 'level'}.get(x, x)
            dims = [io2cf(d) for d in olddims]
        dims = [d for d in dims]  # Why was I excluding time  if d != 'time'
        if olddims != dims:
            if (varkey not in ('latitude', 'longitude') and
                ('PERIM' in dims or
                 ('latitude' in dims and 'longitude' in dims))):
                try:
                    var.coordinates = ' '.join(dims)
                    var.grid_mapping = gridname
                except Exception as e:
                    warn(('coordinates="{0}" and gridmapping="{1}" ' +
                          'not added to variables:\n\t{3}'
                          ).format(' '.join(dims), gridname, varkey, e),
                         category=UserWarning)


def add_ioapi_from_cf(ifileo, coordkeys=[]):
    from ...coordutil import gettimes
    times = gettimes(ifileo)
    jdays = np.array([int(t.strftime('%Y%j')) for t in times])
    itimes = np.array([int(t.strftime('%H%M%S')) for t in times])
    outkeys = [k.ljust(16) for k in ifileo.variables.keys()
               if k not in ('ETFLAG', 'TFLAG') and k not in coordkeys]
    setattr(ifileo, 'VAR-LIST', ''.join(outkeys))
    setattr(ifileo, 'NVARS', len(outkeys))
    ifileo.createDimension('VAR', ifileo.NVARS)
    setattr(ifileo, 'SDATE', jdays[0])
    setattr(ifileo, 'STIME', itimes[0])
    setattr(ifileo, 'EDATE', jdays[-1])
    setattr(ifileo, 'ETIME', itimes[-1])
    tflag = ifileo.createVariable('TFLAG', 'i', ('TSTEP', 'VAR', 'DATE-TIME'))
    tflag[:, :, 0] = jdays[:, None]
    tflag[:, :, 1] = itimes[:, None]
    tflag.units = "<YYYYDDD,HHMMSS>"
    tflag.long_name = "TFLAG           "
    tflag.var_desc = ("Timestep-valid flags: " +
                      " (1) YYYYDDD or (2) HHMMSS").ljust(80)


add_ioapi_from_ioapi = add_ioapi_from_cf


def add_cf_from_ioapi(ifileo, coordkeys=[]):
    try:
        add_lay_coordinates(ifileo)
    except Exception:
        pass
    try:
        add_time_variables(ifileo)
    except Exception:
        pass
    if ifileo.GDTYP in _gdnames:
        add_lcc_coordinates(ifileo)
    else:
        raise TypeError('IOAPI is only aware of LCC (GDTYP=2)')
    ifileo.Conventions = 'CF-1.6'



from __future__ import unicode_literals

__all__ = ['ncf2lateral_boundary']

import numpy as np
from PseudoNetCDF._getwriter import registerwriter

_emiss_hdr_fmt = np.dtype(dict(names=['SPAD', 'name', 'note', 'itzon', 'nspec',
                                      'ibdate', 'btime', 'iedate', 'etime',
                                      'EPAD'],
                               formats=['>i', '(10,4)>S1', '(60,4)>S1', '>i',
                                        '>i', '>i', '>f', '>i', '>f', '>i']))

_grid_hdr_fmt = np.dtype(dict(names=['SPAD', 'plon', 'plat', 'iutm', 'xorg',
                                     'yorg', 'delx', 'dely', 'nx', 'ny', 'nz',
                                     'iproj', 'istag', 'tlat1', 'tlat2',
                                     'rdum5', 'EPAD'],
                              formats=['>i', '>f', '>f', '>i', '>f', '>f',
                                       '>f', '>f', '>i', '>i', '>i', '>i',
                                       '>i', '>f', '>f', '>f', '>i']))

_cell_hdr_fmt = np.dtype(dict(names=['SPAD', 'ione1', 'ione2', 'nx', 'ny',
                                     'EPAD'],
                              formats=['>i', '>i', '>i', '>i', '>i', '>i']))

_time_hdr_fmt = np.dtype(dict(names=['SPAD', 'ibdate', 'btime', 'iedate',
                                     'etime', 'EPAD'],
                              formats=['>i', '>i', '>f', '>i', '>f', '>i']))

_spc_fmt = np.dtype("(10,4)>S1")


def ncf2lateral_boundary(ncffile, outpath):
    emiss_hdr = np.zeros(shape=(1,), dtype=_emiss_hdr_fmt)
    emiss_hdr[0]['name'][:, :] = ' '
    emiss_hdr[0]['name'][:, 0] = np.array(ncffile.NAME, dtype='>c')
    emiss_hdr[0]['note'][:, :] = ' '
    emiss_hdr[0]['note'][:, 0] = np.array(ncffile.NOTE, dtype='>c')
    # gdtype = getattr(ncffile, 'GDTYPE', -999)
    emiss_hdr['itzon'][0] = ncffile.ITZON
    nspec = len(ncffile.dimensions['VAR']) / 4
    emiss_hdr['nspec'] = nspec
    emiss_hdr['ibdate'] = ncffile.SDATE % (ncffile.SDATE / 100000 * 100000)
    emiss_hdr['btime'] = ncffile.STIME / 100.
    EDATE, ETIME = ncffile.variables['TFLAG'][-1, 0, :]
    emiss_hdr['iedate'] = EDATE % (EDATE / 100000 * 100000)
    emiss_hdr['etime'] = ETIME / 10000. + 1.
    emiss_hdr['SPAD'] = _emiss_hdr_fmt.itemsize - 8
    emiss_hdr['EPAD'] = _emiss_hdr_fmt.itemsize - 8

    NCOLS = len(ncffile.dimensions['COL'])
    NROWS = len(ncffile.dimensions['ROW'])
    NLAYS = len(ncffile.dimensions['LAY'])
    grid_hdr = np.zeros(shape=(1,), dtype=_grid_hdr_fmt)
    grid_hdr['SPAD'] = grid_hdr.itemsize - 8
    grid_hdr['plon'] = ncffile.PLON
    grid_hdr['plat'] = ncffile.PLAT
    grid_hdr['iutm'][0] = ncffile.IUTM
    grid_hdr['xorg'] = ncffile.XORIG
    grid_hdr['yorg'] = ncffile.YORIG
    grid_hdr['delx'] = ncffile.XCELL
    grid_hdr['dely'] = ncffile.YCELL
    grid_hdr['nx'] = NCOLS
    grid_hdr['ny'] = NROWS
    grid_hdr['nz'] = NLAYS
    grid_hdr['iproj'] = ncffile.CPROJ
    grid_hdr['tlat1'] = ncffile.TLAT1
    grid_hdr['tlat2'] = ncffile.TLAT2
    grid_hdr['istag'] = ncffile.ISTAG
    grid_hdr['rdum5'] = 0.
    grid_hdr['EPAD'] = grid_hdr.itemsize - 8

    cell_hdr = np.zeros(shape=(1,), dtype=_cell_hdr_fmt)
    cell_hdr['SPAD'] = cell_hdr['EPAD'] = cell_hdr.itemsize - 8
    cell_hdr['ione1'] = 1
    cell_hdr['ione2'] = 1
    cell_hdr['nx'] = grid_hdr['nx']
    cell_hdr['ny'] = grid_hdr['ny']

    time_hdr = np.zeros(
        shape=(len(ncffile.dimensions['TSTEP']),), dtype=_time_hdr_fmt)
    time_hdr['SPAD'] = 16
    time_hdr['EPAD'] = 16
    date, time = ncffile.variables['TFLAG'][:, 0].T
    time = time.astype('>f') / 10000.
    date = date % (date // 100000 * 100000)
    time_hdr['ibdate'] = date
    time_hdr['btime'] = time
    time_hdr['iedate'] = date
    time_hdr['etime'] = time + 1.
    time_hdr['iedate'] += (time_hdr['etime'] // 24).astype('i')
    time_hdr['etime'] -= (time_hdr['etime'] // 24) * 24
    emiss_hdr['ibdate'] = time_hdr['ibdate'][0]
    emiss_hdr['btime'] = time_hdr['btime'][0]
    emiss_hdr['iedate'] = time_hdr['iedate'][-1]
    emiss_hdr['etime'] = time_hdr['etime'][-1]
    emiss_hdr['SPAD'] = _emiss_hdr_fmt.itemsize - 8
    emiss_hdr['EPAD'] = _emiss_hdr_fmt.itemsize - 8

    spcstrfmt = np.dtype("(%d,10,4)>S1" % nspec)
    spc_hdr = np.zeros(shape=(1,), dtype=dict(names=['SPAD1', 'DATA', 'EPAD1'],
                                              formats=['>i', spcstrfmt, '>i']))
    spc_hdr['SPAD1'] = nspec * 40
    spc_hdr['EPAD1'] = nspec * 40
    spc_names = np.array(getattr(ncffile, 'VAR-LIST'),
                         dtype='>c').reshape(-1, 16)[:-1:4, 5:-1].copy()
    spc_hdr[0]['DATA'][:] = ' '
    spc_hdr[0]['DATA'][:, :, 0] = spc_names
    spc_names = spc_names.view('>S10')
    outfile = open(outpath, 'wb')
    emiss_hdr.tofile(outfile)
    grid_hdr.tofile(outfile)
    cell_hdr.tofile(outfile)
    spc_hdr.tofile(outfile)
    for ename, ei in [('WEST', 1), ('EAST', 2), ('SOUTH', 3), ('NORTH', 4)]:
        if hasattr(ncffile, '_boundary_def'):
            ncffile._boundary_def[ename].tofile(outfile)
        else:
            nbcell = dict(WEST=NROWS, EAST=NROWS,
                          SOUTH=NCOLS, NORTH=NCOLS)[ename]
            buf = (nbcell * 4 + 3) * 4
            if ename in ('WEST', 'SOUTH'):
                icell = 2
            elif ename == 'EAST':
                icell = NCOLS - 1
            elif ename == 'NORTH':
                icell = NROWS - 1
            else:
                raise KeyError(
                    'WEST, EAST, SOUTH, or NORTH: received %s' % ename)
            np.array([buf, 1, ei, nbcell, 0, 0, 0, 0] + [icell, 0, 0, 0] *
                     (nbcell - 2) + [0, 0, 0, 0, buf]
                     ).astype('>i').tofile(outfile)
    for di, (d, t) in enumerate(ncffile.variables['TFLAG'][:, 0]):
        tempout = b''
        tempout += time_hdr[di].tobytes()
        for spc_key, spc_name in zip(spc_names, spc_hdr[0]['DATA']):
            for ename, ei in [('WEST', 1), ('EAST', 2),
                              ('SOUTH', 3), ('NORTH', 4)]:
                var = ncffile.variables[ename +
                                        '_' + spc_key[0].decode().strip()]
                data = var[di].astype('>f')
                buf = np.array(4 + 40 + 4 + data.size * 4).astype('>i')
                tempout += buf.tobytes()
                tempout += np.array(1).astype('>i').tobytes()
                tempout += spc_name.tobytes()
                tempout += np.array(ei).astype('>i').tobytes()
                tempout += data.tobytes()
                tempout += buf.tobytes()
        outfile.write(tempout)
    outfile.flush()
    return outfile


registerwriter('camxfiles.lateral_boundary', ncf2lateral_boundary)
registerwriter('lateral_boundary', ncf2lateral_boundary)



__all__ = ['register_cmap', 'get_cmap', 'register_norm', 'get_norm']

from matplotlib.colors import from_levels_and_colors, Normalize
from matplotlib.cm import register_cmap, get_cmap

_registered_norms = {}


def get_norm(name):
    global _registered_norms
    return _registered_norms[name]


def register_norm(name, norm):
    global _registered_norms
    if isinstance(norm, Normalize):
        _registered_norms[name] = norm
    else:
        raise TypeError(
            'cmap should be an instance of Normalize; got ' + str(type(norm)))


def _addusepaaqi():
    aqicolors = ['green', 'yellow', 'orange', 'red', 'purple', 'maroon']
    aqic = aqicolors
    bounds_colors = {}
    bounds_colors[('O3', 'mda8', 'ppbv')] = ([0, 55, 71, 86, 106, 200],
                                             aqic[:5] + ['black'], 'max')
    bounds_colors[('O3', 'mda8', 'ppmv')] = ([0, 0.055, 0.071, 0.086,
                                              0.106, 0.0200],
                                             aqic[:5] + ['black'], 'max')
    bounds_colors[('O3', 'mda1', 'ppbv')] = ([125, 165, 205, 405, 604],
                                             ['white'] + aqic[2:] + ['black'],
                                             'both')
    pmb = [0, 12.1, 35.5, 55.5, 150.5, 250.5, 500.4]
    pmc = aqic + ['black']
    bounds_colors[('PM25', 'a24', 'micrograms/m**3')] = (pmb, pmc, 'max')
    for key, (bnds, colors, extend) in bounds_colors.items():
        print(key)
        name = '_'.join(('usepa', 'aqi',) + key)
        cmap, norm = from_levels_and_colors(bnds, colors, extend=extend)
        register_cmap(name=name, cmap=cmap)
        register_norm(name=name, norm=norm)


_addusepaaqi()


if __name__ == '__main__':
    from matplotlib import pyplot as plt
    import numpy as np
    varkey = 'O3'
    intv = 'mda8'
    unit = 'ppbv'

    colorscale = 'usepa_aqi_' + '_'.join([varkey, intv, unit])
    norm = get_norm(colorscale)
    cmap = get_cmap(colorscale)
    data = np.linspace(25, 200, 16).reshape(4, 4)
    print(np.flipud(data))
    plt.pcolor(data, norm=norm, cmap=cmap)
    plt.savefig('test.png')



from PseudoNetCDF import PseudoNetCDFFile
from .arltime import arl2timea
import numpy as np

_units = dict(trajid='---',
              metgridid='---',
              year='year',
              month='month of year',
              day='day of month',
              hour='hour of day',
              minute='minute of hour',
              forecast_hour='hour in forecast',
              age='hours',
              pressure='hPa',
              theta='K',
              air_temp='K',
              rainfall='mm/h',
              mixdepth='m',
              relhumid='%',
              terr_msl='m',
              sun_flux='W/m**2',)


class arltrajdump(PseudoNetCDFFile):
    @classmethod
    def isMine(cls, path):
        try:
            arltrajdump(path)
            return True
        except Exception:
            return False

    def __init__(self, path):
        self._path = path
        f = self._file = open(path)
        """
        Record #1

        I6 - Number of meteorological grids used in calculation
        """
        nmgrid = self.NMETGRIDS = int(f.readline()[:6].strip())
        """
        Records Loop #2 through the number of grids

        A8 - Meteorological Model identification
        5I6 - Data file starting Year, Month, Day, Hour, Forecast Hour
        """
        metgridlines = [f.readline().strip().split() for i in range(nmgrid)]
        self.METGRIDS = ','.join([l[0] for l in metgridlines])
        self.createDimension('metgrid', nmgrid)
        v = self.createVariable('met_year', 'i', ('metgrid',))
        v.units = 'year'
        v.long_name = 'year'
        v[:] = [int(l[1]) for l in metgridlines]
        v = self.createVariable('met_month', 'i', ('metgrid',))
        v.units = 'month'
        v.long_name = 'month of the year'
        v[:] = [int(l[2]) for l in metgridlines]
        v = self.createVariable('met_day', 'i', ('metgrid',))
        v.units = 'day'
        v.long_name = 'day of the month'
        v[:] = [int(l[3]) for l in metgridlines]
        v = self.createVariable('met_hour', 'i', ('metgrid',))
        v.units = 'hour'
        v.long_name = 'hour of the day (GMT)'
        v[:] = [int(l[4]) for l in metgridlines]
        v = self.createVariable('met_forecast_hour', 'i', ('metgrid',))
        v.units = 'forecast_hour'
        v.long_name = 'hour of the forecast'
        v[:] = [int(l[5]) for l in metgridlines]

        """
        Record #3

        I6 - number of different trajectories in file
        1X,A8 - direction of trajectory calculation (FORWARD, BACKWARD)
        1X,A8 - vertical motion calculation method (OMEGA, THETA, ...)
        """
        ntrajstr, tcalc, vmot = f.readline().strip().split()
        self.TRAJECTORY_CALCULATION, self.VERTMOTION = tcalc, vmot
        ntrajs = self.NTRAJECTORIES = int(ntrajstr)
        """
        Record Loop #4 through the number of different trajectories in file

        4I6 - starting year, month, day, hour
        2F9.3 - starting latitude, longitude
        F8.1 - starting level above ground (meters)
        """
        trajmeta = np.array([f.readline().strip().split()
                             for i in range(ntrajs)], dtype='f')
        self.createDimension('trajectory', ntrajs)
        v = self.createVariable('traj_year', 'i', ('trajectory',))
        v.units = 'year'
        v.long_name = 'year'
        v[:] = trajmeta[:, 0].astype('i')
        v = self.createVariable('traj_month', 'i', ('trajectory',))
        v.units = 'month'
        v.long_name = 'month of the year'
        v[:] = trajmeta[:, 1].astype('i')
        v = self.createVariable('trajectory_day', 'i', ('trajectory',))
        v.units = 'day'
        v.long_name = 'day of the month'
        v[:] = trajmeta[:, 2].astype('i')
        v = self.createVariable('trajectory_hour', 'i', ('trajectory',))
        v.units = 'hour'
        v.long_name = 'hour of the day (GMT)'
        v[:] = trajmeta[:, 3].astype('i')
        v = self.createVariable(
            'trajectory_init_latitude', 'f', ('trajectory',))
        v.units = 'degrees_north'
        v.long_name = 'initial latitude'
        v[:] = trajmeta[:, 4]
        v = self.createVariable(
            'trajectory_init_longitude', 'f', ('trajectory',))
        v.units = 'degrees_east'
        v.long_name = 'initial longitude'
        v[:] = trajmeta[:, 5]
        v = self.createVariable('trajectory_init_height', 'f', ('trajectory',))
        v.units = 'meters agl'
        v.long_name = 'initial altitude'
        v[:] = trajmeta[:, 6]
        # Starting time
        self._starttimes = arl2timea(trajmeta[:, 0], trajmeta[:, 1],
                                     trajmeta[:, 2], trajmeta[:, 3],
                                     trajmeta[:, 3] * 0)
        """
        Record #5

        I6 - number (n) of diagnostic output variables
        n(1X,A8) - label identification of each variable (PRESSURE, THETA, ...)
        """
        diagnostics = f.readline().strip().lower().split()
        ndiag = int(diagnostics[0])
        assert((ndiag + 1) == len(diagnostics))
        """
        Record Loop #6 through the number of hours in the simulation

        I6 - trajectory number
        I6 - meteorological grid number or antecedent trajectory number
        5I6 - year month day hour minute of the point
        I6 - forecast hour at point
        F8.1 - age of the trajectory in hours
        2F9.3 - position latitude and longitude
        1X,F8.1 - position height in meters above ground
        n(1X,F8.1) - n diagnostic output variables (1st to be output
                    is always pressure)
        """
        try:
            import pandas as pd
        except Exception:
            raise ImportError('arltrajdump requires pandas; ' +
                              'install pandas (e.g., pip install pandas)')
        # , parse_dates = ['YEAR MONTH DAY HOUR MINUTE'.split()])
        data = pd.read_csv(f, delimiter='\s+',
                           names=(('trajid metgridid year month day hour ' +
                                   'minute forecast_hour age latitude ' +
                                   'longitude altitude').split() +
                                  diagnostics[1:]))
        mytimes = arl2timea(data['year'], data['month'], data['day'],
                            data['hour'], data['minute'])
        unique_times = np.sort(np.unique(mytimes))
        ntimes = len(unique_times)
        self.createDimension('time', ntimes)
        utraj = data['trajid'].unique()
        mytraj = data['trajid'].values
        # myage = data['age'].values
        trajidx = (utraj[:, None] == mytraj[None, :]).argmax(0)
        timeidx = (unique_times[:, None] == mytimes[None, :]).argmax(0)

        tmpv = np.ma.masked_all((ntimes, ntrajs), dtype='f')
        for k in data.columns:
            v = self.createVariable(
                k, 'f', ('time', 'trajectory'), fill_value=-999.)
            v.long_name = k
            v.units = _units.get(k, 'unknown')
            v[:] = tmpv
            v[timeidx, trajidx] = data[k].values

    def getTimes(self):
        year = (self.variables['year']).max(1).astype('l')
        month = self.variables['month'].max(1).ravel().astype('l')
        day = self.variables['day'].max(1).astype('l')
        hour = self.variables['hour'].max(1).astype('l')
        minute = self.variables['minute'].max(1).astype('l')
        return arl2timea(year, month, day, hour, minute)


if __name__ == '__main__':
    f = arltrajdump('tdump_008')



__all__ = ['PseudoNetCDFDimension']

from collections import OrderedDict


class PseudoNetCDFDimensions(OrderedDict):
    def __repr__(self):
        outs = []
        for k, v in self.items():
            out = '%s = %d' % (k, len(v))
            if v.isunlimited():
                out += ' // UNLIMITED'
            outs.append(out)
        return '\n'.join(outs)


class PseudoNetCDFDimension(object):
    """
    Dimension object responds like that of netcdf4-python
    """

    def __init__(self, group, name, size):
        self._len = int(size)
        self._unlimited = False
        self._name = name

    def isunlimited(self):
        """
        Returns
        -------
        out : boolean
            is the dimension unlimited
        """
        return self._unlimited

    def __len__(self):
        return self._len

    def setunlimited(self, unlimited):
        """
        Parameters
        ----------
        unlimited : boolean
            largely decorative in PseudoNetCDF, but important in persisting
        """
        self._unlimited = unlimited

    def __repr__(self):
        out = object.__repr__(self).replace(
            ' at ', ' (name = %s, len = %d) at' % (self._name, len(self)))
        return out



__doc__ = r"""
Scientific data has dimensions that have physical meaning and values
only have meaning in the context of their units.  This module implements
numpy arrays that are aware of their dimensions trying to vaguely adhere
to the Common Data Model from Unitdata at UCAR.

Each variable has as a property its dimensions names (dimensions).  Further,
each dimension name exists as a property and contains a one dimensional array
of values associated with that dimension.

For the purposes of ease of use, the standard properties of netCDF files
are attached and the arrays implement the Scientific.IO.NetCDF.NetCDFVariable
interfaces.
"""

__all__ = ['PseudoNetCDFFile', 'PseudoNetCDFDimension',
           'PseudoNetCDFVariableConvertUnit',
           'PseudoNetCDFVariable',
           'PseudoNetCDFMaskedVariable',
           'PseudoIOAPIVariable',
           'PseudoNetCDFVariables',
           'Pseudo2NetCDF',
           'reduce_dim',
           'slice_dim',
           'getvarpnc',
           'interpvars',
           'extract',
           'pncbo',
           'seqpncbo',
           'pncexpr',
           'WrapPNC',
           'manglenames',
           'merge',
           'pncrename',
           'removesingleton',
           'splitdim',
           'mask_vals',
           'add_attr',
           'stack_files',
           'extract_from_file',
           'mesh_dim',
           'convolve_dim',
           'get_ncf_object',
           'get_dimension_length']


from .core._files import PseudoNetCDFFile, PseudoNetCDFVariables
from .core._wrapnc import WrapPNC
from .core._dimensions import PseudoNetCDFDimension
from .core._variables import PseudoNetCDFVariable, PseudoNetCDFMaskedVariable
from .core._variables import PseudoIOAPIVariable
from .core._functions import interpvars, extract, mask_vals, slice_dim
from .core._functions import reduce_dim, mesh_dim, pncbo, pncexpr, seqpncbo
from .core._functions import getvarpnc, add_attr, stack_files, convolve_dim
from .core._functions import manglenames, removesingleton, merge
from .core._functions import extract_from_file, pncrename, splitdim
from .core._util import get_ncf_object, get_dimension_length
from .core._transforms import PseudoNetCDFVariableConvertUnit
from PseudoNetCDF.pncgen import Pseudo2NetCDF
